"""Websauna framework initialization routine."""

#
# NOTE: Be careful about not to import anything in this file head that might cause the import of `transaction` package. If Websauna is used with gevent async event loop and gevent monkey patches, the patching order would become incorrect. This is because importlib and setuptools autoscan all top level imports on the application startup and this order cannot be controlled. Thus, to make gevent work you must not import anything dbsession/tranaction related in the package main file to be on the safe side.
#
# This, and the reason that we can use component architecture and replace and disable framework parts in Initializer, are the driving factors of not to do module level imports here.
#

# Standard Library
import logging
import os
import sys
import typing as t
from distutils.version import LooseVersion

# Pyramid
from pyramid.config import Configurator
from pyramid.interfaces import IDebugLogger
from pyramid.interfaces import IRequest
from pyramid.path import DottedNameResolver
from pyramid.settings import asbool
from pyramid.settings import aslist

import pkg_resources

# Websauna
from websauna.utils.autoevent import event_source
from websauna.utils.config.includer import IncludeAwareConfigParser  # noQA


class SanityCheckFailed(Exception):
    """Looks like the application has configuration which would fail to run."""


class RequirementsFailed(Exception):
    """Websauna minimum requirements were not met."""


def check_python_pyramid_requirements() -> bool:
    """Check for Python and Pyramid requirements.

    :return: Boolean indicating if Python and Pyramid requirements are met.
    """
    error_msg = ''
    python_version = sys.version_info
    pyramid_version = tuple(LooseVersion(pkg_resources.get_distribution('pyramid').version).version)
    if python_version < (3, 5, 2):
        error_msg = 'Python 3.5.2 or newer is required to run Websauna.'
    elif pyramid_version < (1, 7):
        error_msg = 'Pyramid version 1.7 or newer required'
    if error_msg:
        raise RequirementsFailed(error_msg)
    return True


def _expandvars(value: t.Any) -> t.Any:
    processed = value
    if isinstance(value, dict):
        processed = expandvars_dict(value)
    elif isinstance(value, (str, bytes)):
        processed = os.path.expandvars(value)
    return processed


def expandvars_dict(settings: dict) -> dict:
    """Expand all environment variables in a settings dictionary.

    ref: http://stackoverflow.com/a/16446566
    :returns: Dictionary with settings
    """
    return {key: _expandvars(value) for key, value in settings.items()}


class Initializer:
    """Initializer is responsible to ramp up the frameworks and subsystems.

    There exist one ``Initializer`` instance which you create in your WSGI application constructor.

    * You subclass the default ``Initializer`` provider by Websauna

    * You override the methods for the parts where you want to customize the default Websauna behavior

    * You also need to include addons and other Pyramid package configurations. This is usually done by calling ``self.config.include("your_pyramid_package")``.

    * You can add your own application specific view initializations, like ``self.config.scan()`` for your application Python modules to register ``@view_config`` directives in those.

    See :py:meth:`websauna.system.Initializer.run` for linear initialization order.

    Aspect-oriented approach with :py:mod:`websauna.utils.aspect.event_source` is used to provide hooks for addons to participate initialization process.

    .. note ::

        We deliberate push imports inside methods, so that it is unlikely we'd have any import time side effects caused by some less elegant solutions, like gevent.
    """

    def __init__(self, global_config: dict, settings: t.Optional[dict]=None):
        """
        :param global_config: Dictionary as passed to WSGI entry point.

        :param settings: DEPRECATED. Extra settings as passed to WSGI entry point. TODO: How to handle these?
        """
        import plaster

        # Check if Python and Pyramid versions are the required
        check_python_pyramid_requirements()

        if not settings:
            config = global_config['__file__']
            if not config.startswith('ws://'):
                config = 'ws://{0}'.format(config)

            loader = plaster.get_loader(config)
            # Read [app] section
            settings = loader.get_settings('app:main')

        #: This is the refer    ence to the config file which started our process. We need to later pass it to Notebook.
        settings["websauna.global_config"] = global_config
        self.global_config = global_config

        #: Reference to Celery app instance
        self.celery = None

        self.settings = expandvars_dict(settings)

        self.config = self.create_configurator()

        self.config.registry.static_asset_policy = self.static_asset_policy = self.create_static_asset_policy()

        #: Flag to tell if we need to do sanity check for redis sessiosn
        self._has_redis_sessions = False

        #: This flag keeps state if the initializer has been run or not.
        self._already_run = False

        # Exposed Websauna features
        self.config.registry.features = set()

        # Secrets file
        self.secrets = self.read_secrets()

    def create_configurator(self) -> Configurator:
        """Create Pyramid Configurator instance."""
        configurator = Configurator(settings=self.settings)

        # This is passed to addons
        configurator.registry.initializer = self
        return configurator

    def create_static_asset_policy(self):
        """Override to have our own static asset policy."""
        from websauna.system.http.static import DefaultStaticAssetPolicy
        return DefaultStaticAssetPolicy(self.config)

    def configure_logging(self):
        """Create and set Pyramid debug logger.

        Please note that most o the logging is configured through the configuration file and that should be the primary way to do it.
        """

        # Extract logging configuration from INI
        # from websauna.utils.configincluder import setup_logging

        # setup_logging(self.global_config["__file__"])

        # Make sure we can target Pyramid router debug messages in logging configuration
        pyramid_debug_logger = logging.getLogger("pyramid_debug")
        self.config.registry.registerUtility(pyramid_debug_logger, IDebugLogger)

    @event_source
    def configure_user_forms(self):
        """Configure forms and schemas used for login and such.

        :ref:`See documentation example <customize-user-form>`.
        """

        from websauna.system.user import interfaces
        from websauna.system.user import schemas
        from websauna.system.user.forms import DefaultUserForm
        from websauna.system.user.forms import DefaultLoginForm
        from websauna.system.user.forms import DefaultRegisterForm

        self.config.registry.registerUtility(schemas.RegisterSchema, interfaces.IRegisterSchema)
        self.config.registry.registerUtility(schemas.LoginSchema, interfaces.ILoginSchema)
        self.config.registry.registerUtility(schemas.ResetPasswordSchema, interfaces.IResetPasswordSchema)
        self.config.registry.registerUtility(schemas.ForgotPasswordSchema, interfaces.IForgotPasswordSchema)

        self.config.registry.registerUtility(DefaultLoginForm, interfaces.ILoginForm)
        self.config.registry.registerUtility(DefaultRegisterForm, interfaces.IRegisterForm)
        self.config.registry.registerUtility(DefaultUserForm, interfaces.IForgotPasswordForm)
        self.config.registry.registerUtility(DefaultUserForm, interfaces.IResetPasswordForm)

    @event_source
    def configure_mailer(self):
        """Configure outgoing email backend and email test views."""
        from pyramid_mailer import IMailer
        from websauna.system.mail.utils import create_mailer

        # TODO: There is upstrean issue in pyramid_mailer, see send_template_mail.
        # mailer gets created again
        mailer = create_mailer(self.config.registry)
        self.config.registry.registerUtility(mailer, IMailer)

        if self.config.registry.settings.get("websauna.sample_html_email", False):
            from websauna.system.mail import views
            self.config.scan(views)
            self.config.add_jinja2_search_path('websauna.system.mail:templates', name='.html')

    @event_source
    def configure_templates(self):
        from websauna.system.core import templatecontext  # noQA
        from websauna.system.core.render import get_on_demand_resource_renderer

        # Jinja 2 templates as .html files
        self.config.include('pyramid_jinja2')
        self.config.add_jinja2_renderer('.html')
        self.config.add_jinja2_renderer('.txt')
        self.config.add_jinja2_renderer('.css')
        self.config.add_jinja2_renderer('.xml')

        self.config.include("websauna.system.core.templatecontext")
        self.config.include("websauna.system.core.vars")

        # Add core templates to the search path
        self.config.add_jinja2_search_path('websauna.system.core:templates', name='.html')
        self.config.add_jinja2_search_path('websauna.system.core:templates', name='.txt')
        self.config.add_jinja2_search_path('websauna.system.core:templates', name='.xml')
        self.config.add_jinja2_search_path('websauna.system.core:templates', name='.css')

        # Add the default resource registry for Deform
        self.config.add_request_method(get_on_demand_resource_renderer, 'on_demand_resource_renderer', reify=True)

    @event_source
    def configure_authentication(self):
        """Set up authentication and authorization policies.

        For more information see Pyramid auth documentation.
        """
        from websauna.system.auth.principals import resolve_principals
        from websauna.system.auth.authentication import get_request_user
        from pyramid.authorization import ACLAuthorizationPolicy
        from websauna.system.auth.policy import SessionAuthenticationPolicy

        authn_policy = SessionAuthenticationPolicy(callback=resolve_principals)
        authz_policy = ACLAuthorizationPolicy()
        self.config.set_authentication_policy(authn_policy)
        self.config.set_authorization_policy(authz_policy)

        # We need to carefully be above TM view, but below exc view so that internal server error page doesn't trigger session authentication that accesses the database
        self.config.add_tween("websauna.system.auth.tweens.SessionInvalidationTweenFactory", under="pyramid_tm.tm_tween_factory")

        # Grab incoming auth details changed events
        from websauna.system.auth import subscribers
        self.config.scan(subscribers)

        # Experimental support for transaction aware properties
        try:
            from pyramid_tm.reify import transaction_aware_reify
            self.config.add_request_method(
                callable=transaction_aware_reify(self.config, get_request_user),
                name="user",
                property=True,
                reify=False)
        except ImportError:
            self.config.add_request_method(get_request_user, 'user', reify=True)

    @event_source
    def configure_panels(self):
        self.config.include('pyramid_layout')

    @event_source
    def configure_federated_login(self):
        """Configure federated authentication (OAuth).

        Set up Authomatic login services.

        Read enabled federated authentication methods from the configuration file.
        """

        # TODO: Refactor this to separate functions, not implementation is not very clean

        import authomatic
        from websauna.system.user.interfaces import IAuthomatic, ISocialLoginMapper, IOAuthLoginService
        from websauna.system.user.oauthloginservice import DefaultOAuthLoginService

        settings = self.settings
        secrets = self.secrets

        self.config.add_route('login_social', '/login/{provider_name}')

        social_logins = aslist(settings.get("websauna.social_logins", ""))

        if not social_logins:
            return

        authomatic_config = {}

        authomatic_secret = secrets["authomatic.secret"]

        resolver = DottedNameResolver()

        # Quick helper to access settings
        def xget(section, key):
            value = secrets.get(section + "." + key)
            assert value is not None, "Missing secret settings for [{}]: {}".format(section, key)
            return value

        for login in social_logins:

            authomatic_config[login] = {}
            authomatic_config[login]["consumer_key"] = xget(login, "consumer_key")
            authomatic_config[login]["consumer_secret"] = xget(login, "consumer_secret")
            authomatic_config[login]["scope"] = aslist(xget(login, "scope"))

            # TODO: Class is not a real secret, think smarter way to do this
            authomatic_config[login]["class_"] = resolver.resolve(xget(login, "class"))

            # Construct social login mapper
            mapper_class = xget(login, "mapper")
            if mapper_class:
                mapper_class = resolver.resolve(mapper_class)
                mapper = mapper_class(self.config.registry, login)
                self.config.registry.registerUtility(mapper, ISocialLoginMapper, name=login)

        # Store instance

        # Pass explicitly a logger so that we can control the log level
        logger = logging.getLogger("authomatic")

        instance = authomatic.Authomatic(config=authomatic_config, secret=authomatic_secret, logger=logger)
        self.config.registry.registerUtility(instance, IAuthomatic)

        self.config.registry.registerAdapter(factory=DefaultOAuthLoginService, required=(IRequest,), provided=IOAuthLoginService)

    @event_source
    def configure_database(self):
        """Configure database.

        * Set up base model

        * Set up mechanism to create database session for requests

        * Set up transaction machinery

        Calls py:func:`websauna.system.model.meta.includeme`.

        """
        from .model.meta import create_transaction_manager_aware_dbsession
        from .model.interfaces import ISQLAlchemySessionFactory

        self.config.include("pyramid_retry")
        self.config.include("pyramid_tm")
        self.config.include(".model.meta")

        self.config.registry.registerAdapter(factory=create_transaction_manager_aware_dbsession, required=(IRequest,), provided=ISQLAlchemySessionFactory)

    @event_source
    def configure_redis(self):
        """Configure Redis connection pool."""
        from websauna.system.core import redis
        self.config.registry.redis = redis.create_redis(self.config.registry)

        # Add logging events
        self.config.scan(redis)

    @event_source
    def configure_instrumented_models(self):
        """Configure models from third party addons and dynamic SQLAlchemy fields which need access to the configuration.

        Third party addons might need references to configurable models which are not available at the import time. One of these models is user - you can supply your own user model. However third party addon models might want to build foreign key relationships to this model. Thus, ``configure_instrumented_models()`` is an initialization step which is called when database setup is half way there and you want to throw in some extra models in.

        This exposes ``Configurator`` to SQLAlchemy through ``websauna.system.model.meta.Base.metadata.pyramid_config`` variable.
        """
        # Expose Pyramid configuration to classes
        from websauna.system.model.meta import Base
        Base.metadata.pyramid_config = self.config

    @event_source
    def configure_error_views(self):

        settings = self.settings

        # Forbidden view overrides helpful auth debug error messages,
        # so pull in only when really needed

        if not asbool(settings["pyramid.debug_authorization"]):
            from websauna.system.core.views import forbidden
            self.config.scan(forbidden)

        if not asbool(settings["pyramid.debug_notfound"]):
            from websauna.system.core.views import notfound
            self.config.scan(notfound)

        # Internal server error page must be only activated in the production mode, as it clashes with pyramid_debugtoolbar, as both handle uncaught exceptions
        has_debug_toolbar = "pyramid_debugtoolbar" in aslist(settings.get("pyramid.includes", []))
        debug_toolbar_enabled = has_debug_toolbar and asbool(settings.get("debugtoolbar.enabled", True))

        if not debug_toolbar_enabled:
            from websauna.system.core.views import internalservererror
            self.config.scan(internalservererror)

        if settings.get("websauna.error_test_trigger", False):
            from websauna.system.core.views import errortrigger
            self.config.scan(errortrigger)
            self.config.add_route('error_trigger', '/error-trigger')

        from websauna.system.core.views import badcsrftoken
        self.config.scan(badcsrftoken)

    @event_source
    def configure_root(self):
        """Root object defines permissions for route URLs which have not their own traversing context.

        http://pyramid-tutorials.readthedocs.org/en/latest/getting_started/10-security/
        """
        from websauna.system.core.root import Root
        self.config.set_root_factory(Root.root_factory)

    @event_source
    def configure_views(self):
        """Configure views for your application."""

    @event_source
    def configure_sitemap(self):
        """Configure sitemap generation for your site.

        By default this is not configured and nothing is done.
        """

    @event_source
    def configure_static(self):
        """Configure static asset views.

        By default we serve only core Websauna assets. Override this to add more static asset declarations to your app.
        """
        self.static_asset_policy.add_static_view('websauna-static', 'websauna.system:static')

    @event_source
    def configure_sessions(self):
        """Configure session storage."""

        from websauna.system.core.session import set_creation_time_aware_session_factory

        session_secret = self.secrets["session.secret"]

        # TODO: Make more boilerplate here so that we pass secret in more sane way
        self.config.registry.settings["redis.sessions.secret"] = session_secret
        self.config.include("pyramid_redis_sessions")

        # Set a flag to perform Redis session check later and prevent web server start if Redis is down
        self._has_redis_sessions = True

        set_creation_time_aware_session_factory(self.config)

    @event_source
    def configure_admin(self):
        """Configure admin ux.

        Register templates and views for admin interface.
        """
        from websauna.system.admin import subscribers
        from websauna.system.admin import views
        from websauna.system.admin.admin import Admin
        from websauna.system.admin.interfaces import IAdmin
        from websauna.system.admin.modeladmin import configure_model_admin
        from websauna.system.admin.utils import get_admin

        # Register default Admin provider
        config = self.config
        config.registry.registerUtility(Admin, IAdmin)

        # Set up model lookup
        configure_model_admin(config)

        config.add_jinja2_search_path('websauna.system.admin:templates', name='.html')
        config.add_jinja2_search_path('websauna.system.admin:templates', name='.txt')

        config.add_route('admin_home', '/admin/', factory="websauna.system.admin.utils.get_admin")
        config.add_route('admin', "/admin/*traverse", factory="websauna.system.admin.utils.get_admin")

        config.add_panel('websauna.system.admin.views.default_model_admin_panel')
        config.scan(views)
        config.scan(subscribers)

        # Add request.admin variable
        self.config.add_request_method(get_admin, 'admin', reify=True)

    def configure_csrf(self):
        """Configure cross-site request forgery subsystem."""
        self.config.set_default_csrf_options(require_csrf=True)

    @event_source
    def configure_forms(self):
        """Configure subsystems for rendering Deform forms.

        * Deform templates

        * Deform JS and CSS

        * CSRf view mapper
        """
        from websauna.system.form.resources import DefaultFormResources
        from websauna.system.form.interfaces import IFormResources
        from websauna.system.form.deform import configure_zpt_renderer

        # Make Deform widgets aware of our widget template paths
        configure_zpt_renderer(["websauna.system.form:templates/deform"])

        # Include Deform JS and CSS to static serving
        self.static_asset_policy.add_static_view('deform-static', 'deform:static')

        # Overrides for Deform 2 stock JS and CSS
        default_form_resources = DefaultFormResources()
        self.config.registry.registerUtility(default_form_resources, IFormResources)

    @event_source
    def configure_crud(self):
        """CRUD templates and views."""
        # Add our template to search path
        self.config.add_jinja2_search_path('websauna.system.crud:templates', name='.html')
        self.config.add_jinja2_search_path('websauna.system.crud:templates', name='.txt')

        from websauna.system.crud import views
        self.config.scan(views)

    @event_source
    def configure_models(self):
        """Configure all models from your application.

        Import related model modules and scan them.

        Importing anything with :py:class:`websauna.system.model.meta.Base` base class registers this model to an SQLAlchemy model registry for migrations.
        """
        pass

    @event_source
    def configure_user_models(self):
        """Plug in user models.

        This initialization step connects chosen user model to SQLAlchemy model Base. Also set up :py:class:`websauna.system.user.usermixin.SiteCreator` logic - what happens when the first user logs in.
        """
        from websauna.system.model.meta import Base
        from websauna.system.model.utils import attach_model_to_base
        from websauna.system.user import models
        from websauna.system.user.interfaces import IActivationModel
        from websauna.system.user.interfaces import IGroupModel
        from websauna.system.user.interfaces import ISiteCreator
        from websauna.system.user.interfaces import IUserModel
        from websauna.system.user.interfaces import IUserRegistry
        from websauna.system.user.usermixin import SiteCreator
        from websauna.system.user.userregistry import DefaultEmailBasedUserRegistry

        attach_model_to_base(models.User, Base)
        attach_model_to_base(models.Group, Base)
        attach_model_to_base(models.Activation, Base)
        attach_model_to_base(models.UserGroup, Base)

        # Mark active user and group class
        registry = self.config.registry
        registry.registerUtility(models.User, IUserModel)
        registry.registerUtility(models.Group, IGroupModel)
        registry.registerUtility(models.Activation, IActivationModel)

        site_creator = SiteCreator()
        registry.registerUtility(site_creator, ISiteCreator)

        # Which user registry we are using
        registry.registerAdapter(factory=DefaultEmailBasedUserRegistry, required=(IRequest,), provided=IUserRegistry)

    @event_source
    def configure_user(self):
        """Configure user model, sign in and sign up subsystem.

        * User services

        * Sign in and sign up templates and views

        * User events
        """
        from websauna.system.user import subscribers
        from websauna.system.user import views
        from websauna.system.user.credentialactivityservice import DefaultCredentialActivityService
        from websauna.system.user.interfaces import ICredentialActivityService
        from websauna.system.user.interfaces import ILoginService
        from websauna.system.user.interfaces import IRegistrationService
        from websauna.system.user.loginservice import DefaultLoginService
        from websauna.system.user.registrationservice import DefaultRegistrationService

        # Set up login service
        registry = self.config.registry
        registry.registerAdapter(factory=DefaultLoginService, required=(IRequest,), provided=ILoginService)
        registry.registerAdapter(factory=DefaultCredentialActivityService, required=(IRequest,), provided=ICredentialActivityService)
        registry.registerAdapter(factory=DefaultRegistrationService, required=(IRequest,), provided=IRegistrationService)

        self.config.add_jinja2_search_path('websauna.system.user:templates', name='.html')
        self.config.add_jinja2_search_path('websauna.system.user:templates', name='.txt')

        self.config.scan(subscribers)
        self.config.scan(views)
        self.config.add_route('waiting_for_activation', '/waiting-for-activation')
        self.config.add_route('registration_complete', '/registration-complete')
        self.config.add_route('login', '/login')
        self.config.add_route('logout', '/logout')
        self.config.add_route('forgot_password', '/forgot-password')
        self.config.add_route('reset_password', '/reset-password/{code}')
        self.config.add_route('register', '/register')
        self.config.add_route('activate', '/activate/{code}')

    @event_source
    def configure_password(self):
        """Configure system password hashing solution.

        By default use Argon 2

        * https://github.com/hynek/argon2_cffi

        For more information see :py:mod:`websauna.system.user.password`
        """
        from websauna.system.user.password import Argon2Hasher
        from websauna.system.user.interfaces import IPasswordHasher

        hasher = Argon2Hasher()

        registry = self.config.registry
        registry.registerUtility(hasher, IPasswordHasher)

    @event_source
    def configure_model_admins(self):
        import websauna.system.user.admins
        import websauna.system.user.adminviews
        self.config.scan(websauna.system.user.admins)
        self.config.scan(websauna.system.user.adminviews)

    @event_source
    def configure_notebook(self):
        """Setup pyramid_notebook integration."""

        # Check if we have IPython installed
        try:
            pkg_resources.get_distribution('IPython[notebook]')
        except pkg_resources.DistributionNotFound:
            return

        try:
            import websauna.system.notebook.views
            import websauna.system.notebook.adminviews
            import websauna.system.notebook.subscribers
        except ImportError:
            # Have installed IPython[Notebook], but not pyramid_notebook
            return

        self.config.add_route('admin_shell', '/notebook/admin-shell')
        self.config.add_route('shutdown_notebook', '/notebook/shutdown')
        self.config.add_route('notebook_proxy', '/notebook/*remainder')
        self.config.scan(websauna.system.notebook.views)
        self.config.scan(websauna.system.notebook.adminviews)
        self.config.scan(websauna.system.notebook.subscribers)
        self.config.registry.features.add("notebook")

    @event_source
    def configure_tasks(self):
        """Scan all Python modules with asynchoronous and periodic tasks to be imported."""
        try:
            import celery  # noQA
        except ImportError as e:
            # Celery not installed as optional dependency
            return

        # Importing the task is enough to add it to Celerybeat working list
        from websauna.system.devop import tasks  # noQA
        self.config.scan(tasks)

    @event_source
    def configure_tweens(self):
        """Configure tweens."""

    def read_secrets(self) -> dict:
        """Read secrets configuration file.

        Stores API keys, such.
        """
        # Secret configuration diretives
        from websauna.utils.secrets import read_ini_secrets
        from websauna.system.core.interfaces import ISecrets

        settings = self.settings

        secrets_file = settings.get("websauna.secrets_file")
        if not secrets_file:
            return {}

        strict = asbool(settings.get("websauna.secrets_strict", True))

        _secrets = read_ini_secrets(secrets_file, strict=strict)
        self.config.registry.registerUtility(_secrets, ISecrets)
        secret_settings = {k.replace('app:main.', ''): v for k, v in _secrets.items() if k.startswith('app:main.')}
        settings.update(secret_settings)
        self.config.registry.settings.update(secret_settings)
        return _secrets

    def include_addons(self):
        """Override this method to include Websauna addons for your app.

        Websauna addons are created with ``websauna_addon`` scaffold.

        By default do nothing.
        """

    def run(self):
        """Run the initialization and prepare Pyramid subsystems.

        This is the main entry for ramping up a Websauna application.
        We go through various subsystem inits.
        """

        # Avoid running the initializer twice. This might happen e.g. due to a bad testing set up where the initializer creation is not scoped properly. E.g. Jinja template engine will get very confused.
        assert not self._already_run, "Attempted to run initializer twice. Please avoid double initialization as it will lead to problems."

        self.configure_logging()

        # Configure addons before anything else, so we can override bits from addon, like template lookup paths, later easily
        self.include_addons()

        # Serving
        self.configure_templates()
        self.configure_static()

        # Authentication and authorization
        # (Must be before any views are included)
        self.configure_authentication()

        # Forms
        self.configure_csrf()
        self.configure_forms()
        self.configure_crud()

        # Email
        self.configure_mailer()

        # Timed tasks
        self.configure_tasks()

        # Core view and layout related
        self.configure_root()
        self.configure_error_views()
        self.configure_views()
        self.configure_panels()
        self.configure_sitemap()
        self.configure_tweens()

        # Website administration
        self.configure_admin()

        # Addon models
        self.configure_models()

        # Redis (preferably before sessions)
        self.configure_redis()

        # Sessions and users
        self.configure_sessions()
        self.configure_user()
        self.configure_user_forms()
        self.configure_user_models()
        self.configure_password()
        self.configure_federated_login()

        # Configure web shell
        self.configure_notebook()

        # Database and models
        self.configure_instrumented_models()
        self.configure_model_admins()
        self.configure_database()

        # Tests can pass us some extra initialization work on ad hoc
        extra_init = self.global_config.get("extra_init")
        if extra_init:
            resolver = DottedNameResolver()
            extra_init = resolver.resolve(extra_init)
            extra_init(self)

        self._already_run = True

    @event_source
    def sanity_check(self):
        """Perform post-initialization sanity checks.

        This is run on every startup to check that the database table schema matches our model definitions. If there are un-run migrations this will bail out and do not let the problem to escalate later.

        See also: :ref:`websauna.sanity_check`.

        """
        import sqlalchemy.exc
        from websauna.system.model import sanitycheck
        from websauna.system.model.meta import Base
        from websauna.system.model.meta import create_dbsession
        from websauna.system.core import redis

        dbsession = create_dbsession(self.config.registry)

        db_connection_string = self.config.registry.settings.get("sqlalchemy.url")

        try:
            if not sanitycheck.is_sane_database(Base, dbsession):
                raise SanityCheckFailed("The database sanity check failed. Check log for details.")
        except sqlalchemy.exc.OperationalError as e:
            raise SanityCheckFailed("The database {} is not responding.\nMake sure the database is running on your local computer or correctly configured in settings INI file.\nFor more information see https://websauna.org/docs/tutorials/gettingstarted/tutorial_02.html.".format(db_connection_string)) from e

        dbsession.close()

        if self._has_redis_sessions:
            if not redis.is_sane_redis(self.config):
                raise SanityCheckFailed("Could not connect to Redis server.\nWebsauna is configured to use Redis server for session data.\nIt cannot start up without a running Redis server.\nPlease consult your operating system community how to install and start a Redis server.")

    @event_source
    def wrap_wsgi_app(self, app):
        """Perform any necessary WSGI application wrapping.

        Wrap WSGI application to another WSGI application e.g. for the monitoring support. By default support New Relic.
        """

        # TODO: Make this plugin
        # if "NEW_RELIC_CONFIG_FILE" in os.environ:
        # Wrap for New Relic
        # libgcc_s.so.1 must be installed for pthread_cancel to work
        # import newrelic.agent
        # return newrelic.agent.wsgi_application()(app)

        return app

    def make_wsgi_app(self, sanity_check=True):
        """Create WSGI application from the current setup.

        :param sanity_check: True if perform post-initialization sanity checks.
        :return: WSGI application
        """
        app = self.config.make_wsgi_app()
        # Carry the initializer around so we can access it in tests

        app.initializer = self

        if "sanity_check" in self.global_config:
            # Command line scripts can override this when calling bootstrap()
            sanity_check = asbool(self.global_config["sanity_check"])
        else:
            sanity_check = asbool(self.settings.get("websauna.sanity_check", True))

        if sanity_check:
            self.sanity_check()

        app = self.wrap_wsgi_app(app)

        return app

    @classmethod
    def bootstrap(cls, global_config):
        """Use as WSGI entry point.

        :return:
        """
        init = cls(global_config)
        init.run()
        wsgi_app = init.make_wsgi_app()
        return wsgi_app


def get_init(global_config, settings, init_cls=None) -> Initializer:
    """Get Initializer class instance for WSGI-like app.

    TODO: Deprecated. Use Pyramid's ``bootstrap()`` instead.

    Reads reference to the initializer from settings, resolves it and creates the initializer instance.

    Example 1::

        config_uri = argv[1]
        init = get_init(dict(__file__=config_uri), settings)

    :param global_config: Global config dictionary, having __file__ entry as given by Paster

    :param settings: Settings dictionary

    :param init_cls: Explicitly give the Initializer class to use, otherwise read ``websauna.init`` settings.
    """

    assert "websauna.init" in settings, "You must have websauna.init setting pointing to your Initializer class"

    assert "__file__" in global_config

    if not init_cls:
        init_cls = settings.get("websauna.init")
        if not init_cls:
            raise RuntimeError("INI file lacks websauna.init option")
        resolver = DottedNameResolver()
        init_cls = resolver.resolve(init_cls)
    init = init_cls(global_config, settings)
    return init


class DemoInitializer(Initializer):
    """Websauna uses internally to run tests, development server."""

    def configure_views(self):
        # Add home view required to run functional test.
        # This view is never exposed to the applications
        # built upon Websauna.
        super(DemoInitializer, self).configure_views()
        from websauna.system.core.views import home
        self.config.add_route('home', '/')
        self.config.scan(home)


def main(global_config, **settings):
    """Entry point for creating a Pyramid WSGI application."""

    init = DemoInitializer(global_config)
    init.run()
    wsgi_app = init.make_wsgi_app()
    return wsgi_app



"""HTTP header helper utilities."""


# https://github.com/pypa/warehouse/blob/master/warehouse/cache/http.py
def add_vary_callback_if_cookie(*varies):
    """Add vary: cookie header to all session responses.

    Prevent downstream web serves to accidentally cache session set-cookie reponses,
    potentially resulting to session leakage.
    """
    def inner(request, response):
        vary = set(response.vary if response.vary is not None else [])
        vary |= set(varies)
        response.vary = vary
    return inner



"""Extra SQLAlchemy column declarations.

We have some sqlalchemy_utils aliasing here intenrally. In the future expect those aliases to be dropped in the favor of dropping sqlalchemy_utils dependency.
"""

# Standard Library
import datetime

# SQLAlchemy
from sqlalchemy import DateTime
from sqlalchemy import processors
from sqlalchemy import types
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import INET as _INET
from sqlalchemy.dialects.postgresql import JSONB as _JSONB
from sqlalchemy.dialects.sqlite import DATETIME as DATETIME_
from sqlalchemy.exc import UnsupportedCompilationError
from sqlalchemy_utils.types.ip_address import IPAddressType
from sqlalchemy_utils.types.json import JSONType
from sqlalchemy_utils.types.uuid import UUIDType

from .json import json_serializer


class UTCDateTime(DateTime):
    """An SQLAlchemy DateTime column that explicitly uses timezone aware dates and only accepts UTC."""

    def __init__(self, *args, **kwargs):
        # If there is an explicit timezone we accept UTC only
        if "timezone" in kwargs:
            if kwargs["timezone"] not in (datetime.timezone.utc, True):
                raise ValueError(
                    "Only 'datetime.timezone.utc' or True accepted"
                    " as timezone for '{}'".format(self.__class__.__name__)
                )

        kwargs = kwargs.copy()
        # Using an explict 'True' ensures no false positives
        # on alembic migrations due to failed equality test.
        # (issue #162)
        kwargs["timezone"] = True
        super(UTCDateTime, self).__init__(**kwargs)

    def _dialect_info(self, dialect):
        if dialect.name == "sqlite":
            # Becase SQLite does not support datetimes, we need to explicitly tell here to use our super duper DATETIME() hack subclass that hacks in timezone
            return {"impl": SQLITEDATETIME()}
        else:
            return super(UTCDateTime, self)._dialect_info(dialect)


class JSONB(JSONType):
    """Generic JSONB type.

    Falls back to the native PostgreSQL JSONB type.
    """

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            # Use the native JSON type.
            return dialect.type_descriptor(_JSONB())
        else:
            return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        if dialect.name == 'postgresql':
            return value
        if value is not None:
            # Correctly process NestedMutationDict
            value = json_serializer(value)
        return value


class INET(IPAddressType):
    """Generic IPv4/IPv6 type.

    Falls back to native PostgreSQL INET implementation.

    The resulting object is either :py:class:`ipaddress.IPv4Address` or :py:class:`ipaddress.IPv6Address`.
    """

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            # Fallback to native PSQL INET
            return dialect.type_descriptor(_INET())
        else:
            return dialect.type_descriptor(self.impl)

    def __str__(self):
        """Printable info for this type.

        Override to avoid issues with logging raising an UnsupportedCompilationError here.
        :return: Representation of this type.
        """
        try:
            repr = str(self.compile())
        except UnsupportedCompilationError:
            dialect = self.impl._default_dialect()
            repr = str(self.compile(dialect=dialect))
        return repr


class UUID(UUIDType):
    """Generic UUID type.

    Falls back to native PostgreSQL UUID implementation.

    The resulting object is :py:class:`uuid.UUID`.
    """

    #: We force PSQL implementation by default here so that Alembic migration scripts don't do a column with unnecessary length attribute: sa.Column('uuid', websauna.system.model.columns.UUID(length=16), nullable=True),
    impl = postgresql.UUID()

    def __init__(self, as_uuid: bool=True):
        super(UUID, self).__init__(binary=True, native=True)

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql' and self.native:
            # Use the native UUID type.
            return dialect.type_descriptor(postgresql.UUID())
        else:
            # Fallback to either a BINARY or a CHAR.
            kind = types.BINARY(16)
            return dialect.type_descriptor(kind)


class SQLITEDATETIME(DATETIME_):
    """Timezone aware datetime support for SQLite.

    This is implementation used for UTCDateTime.
    """

    @staticmethod
    def process(value):
        dt = processors.str_to_datetime(value)
        if dt:
            # Returns naive datetime, force it to UTC
            return dt.replace(tzinfo=datetime.timezone.utc)
        return dt

    def result_processor(self, dialect, coltype):
        return SQLITEDATETIME.process


# Don't expose sqlalchemy_utils internals as they may go away
__all__ = [
    "UTCDateTime",
    "UUID",
    "JSONB",
    "INET",
]



"""Test CSRF functionality as functional tests.

.. note ::

    These tests are leftovers from < Pyramid 1.7 when Pyramid did not have flexible CSRF protection and Websauna provided its own mechanism.

"""
# Pyramid
from pyramid import testing
from pyramid.exceptions import BadCSRFToken
from pyramid.testing import DummySession

import pytest
from webtest import TestApp

from . import csrfsamples


@pytest.fixture()
def csrf_app(request):
    """py.test fixture to set up a dummy app for CSRF testing.

    :param request: pytest's FixtureRequest (internal class, cannot be hinted on a signature)
    """

    session = DummySession()

    config = testing.setUp()
    config.set_default_csrf_options(require_csrf=True)
    config.add_route("home", "/")
    config.add_route("csrf_sample", "/csrf_sample")
    config.add_route("csrf_exempt_sample", "/csrf_exempt_sample")
    config.add_route("csrf_exempt_sample_context", "/csrf_exempt_sample_context")
    config.add_route("csrf_sample_double_argument", "/csrf_sample_double_argument/{arg}")
    config.add_route("csrf_exempt_sample_double_argument", "/csrf_exempt_sample_double_argument/{arg}")
    config.scan(csrfsamples)

    # We need sessions in order to use CSRF feature

    def dummy_session_factory(secret):
        # Return the same session over and over again
        return session

    config.set_session_factory(dummy_session_factory)

    def teardown():
        testing.tearDown()

    app = TestApp(config.make_wsgi_app())
    # Expose session data for tests to read
    app.session = session
    return app


@pytest.fixture()
def session(request, csrf_app):
    return csrf_app.session


def test_csrf_by_default(csrf_app: TestApp, session: DummySession):
    """CSRF goes throgh if we have a proper token."""

    resp = csrf_app.post("/csrf_sample", {"csrf_token": session.get_csrf_token()})
    assert resp.status_code == 200


def test_csrf_by_default_fail(csrf_app: TestApp, session: DummySession):
    """CSRF error is raised by default if we try to POST to a view and we don't have token."""

    with pytest.raises(BadCSRFToken):
        csrf_app.post("/csrf_sample")


def test_csrf_exempt(csrf_app: TestApp, session: DummySession):
    """Decorated views don't have automatic CSRF check."""

    resp = csrf_app.post("/csrf_exempt_sample")
    assert resp.status_code == 200

    resp = csrf_app.post("/csrf_exempt_sample_context")
    assert resp.status_code == 200



"""Default login service implementation."""
# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.response import Response
from pyramid.security import forget
from pyramid.security import remember
from pyramid.settings import asbool
from zope.interface import implementer

# Websauna
from websauna.system.core import messages
from websauna.system.core.route import get_config_route
from websauna.system.http import Request
from websauna.system.user.interfaces import ILoginService
from websauna.system.user.interfaces import IUser
from websauna.system.user.usermixin import UserMixin
from websauna.system.user.utils import get_user_registry
from websauna.utils.time import now

from . import events
from .interfaces import AuthenticationFailure


@implementer(ILoginService)
class DefaultLoginService:
    """A login service which tries to authenticate with email and username against the current user registry.

    Login service must know details about user implementation and user registry abstraction is not enough.
    """

    def __init__(self, request: Request):
        """Initialize LoginService.

        :param request: Pyramid Request.
        """
        self.request = request

    def update_login_data(self, user: IUser):
        """Update last_login_at and last_login_ip on User object.

        If this is the User first login, trigger FirstLogin event.

        :param user: User object.
        """
        request = self.request
        if not user.last_login_at:
            e = events.FirstLogin(request, user)
            request.registry.notify(e)

        # Update user security details
        user.last_login_at = now()
        user.last_login_ip = request.client_addr

    def check_credentials(self, username: str, password: str) -> UserMixin:
        """Check if the user password matches.

        * First try username + password
        + Then try with email + password

        :param username: username or email
        :param password:
        :raise websauna.system.user.interfaces.AuthenticationFailure: On login problem.
        :return: User object which was picked
        """
        request = self.request
        settings = request.registry.settings
        allow_email_auth = settings.get('websauna.allow_email_auth', True)

        # Check login with username
        user_registry = get_user_registry(request)
        user = user_registry.get_authenticated_user_by_username(username, password)

        # Check login with email
        if allow_email_auth and not user:
            user = user_registry.get_authenticated_user_by_email(username, password)

        if not user:
            raise AuthenticationFailure('Invalid username or password.')

        return user

    def greet_user(self, user: IUser):
        """Allow easy overriding of a welcome message.

        :param user: User object.
        """
        messages.add(self.request, kind="success", msg="You are now logged in.", msg_id="msg-you-are-logged-in")

    def do_post_login_actions(self, user: IUser, headers: dict, location: str=None) -> Response:
        """What happens after a successful login.

        Override this to customize e.g. where the user lands.

        :param user: User object.
        :param headers: Dictionary with headers to be added to the HTTPFound response.
        :param location: URL to redirect the user to.
        :return: Redirection to location.
        """
        request = self.request

        if not location:
            location = get_config_route(request, 'websauna.login_redirect')

        self.greet_user(user)

        self.update_login_data(user)

        e = events.Login(request, user)
        request.registry.notify(e)

        return HTTPFound(location=location, headers=headers)

    def authenticate_user(self, user: IUser, login_source: str, location: str=None) -> Response:
        """Make the current session logged in session for this particular user.

        How to authenticate user using the login service (assuming you have done password match or related yourself):

        .. code-block:: python

            from websauna.system.user.utils import get_login_service

            def my_view(request):

                # load user model instance from database
                # user = ...

                login_service = get_login_service(request)
                response = login_service.authenticate_user(user, "my-login-source")

        :param user: User object.
        :param login_source: Source of this login.
        :param location: Location to redirect the user to.
        :raise AuthenticationFailure: If login cannot proceed due to disabled user account, etc.
        :return: HTTPResponse what should happen as post-login action
        """
        request = self.request
        settings = request.registry.settings

        require_activation = asbool(settings.get('websauna.require_activation', True))
        allow_inactive_login = asbool(settings.get('websauna.allow_inactive_login', False))

        if (not allow_inactive_login) and require_activation and (not user.is_activated()):
            raise AuthenticationFailure('Your account is not active, please check your e-mail. If your account activation email as expired please request a password reset.')

        if not user.can_login():
            raise AuthenticationFailure('This user account cannot log in at the moment.')

        user_registry = get_user_registry(request)
        token = user_registry.get_session_token(user)
        headers = remember(request, token)
        # assert headers, "Authentication backend did not give us any session headers"

        return self.do_post_login_actions(user, headers, location)

    def authenticate_credentials(self, username: str, password: str, login_source: str, location: str=None) -> Response:
        """Try logging in the user with username and password.

        This is called after the user credentials have been validated, after sign up when direct sign in after sign up is in use or after successful federated authentication.

        Sets the auth cookies and redirects to a post login page, which defaults to a view named 'index'.

        Fills in user last login time and IP data..

        :param username: Username.
        :param password: User password.
        :param login_source: Source of this login attempt.
        :param location: Override the redirect page. If none use ``websauna.login_redirect``. TODO - to be changed.
        :raise: AuthenticationError
        :return: HTTPResponse what should happen as post-login action
        """
        # See that our user model matches one we expect from the configuration
        user = self.check_credentials(username, password)
        return self.authenticate_user(user, login_source, location)

    def logout(self, location: str=None) -> Response:
        """Log out user from the site.

        * Terminate session
        * Show logged out message
        * Redirect the user to post login page

        :param location: Override the redirect page. If none use ``websauna.login_redirect``. TODO - to be changed.
        :return: HTTPFound to location.
        """
        # TODO: Horus might go
        request = self.request
        logout_redirect_view = get_config_route(request, 'websauna.logout_redirect')
        location = location or logout_redirect_view

        messages.add(request, msg="You are now logged out.", kind="success", msg_id="msg-logged-out")
        headers = forget(request)
        return HTTPFound(location=location, headers=headers)



# Standard Library
import typing as t

# Pyramid
from pyramid.interfaces import IRequest
from pyramid.registry import Registry
from pyramid.request import Request
from pyramid.request import apply_request_extensions
from transaction import TransactionManager

# SQLAlchemy
from sqlalchemy.orm import Session

# Websauna
from websauna.system.model.meta import create_dbsession
from websauna.system.model.meta import create_transaction_manager_aware_dbsession  # noQA


def make_routable_request(dbsession: t.Optional[Session]=None, registry: t.Optional[Registry]=None, path='/') -> IRequest:
    """Creates a dummy request that has route_url and other routing methods.

    As this request does not get HTTP hostname and such stuff from WSGI environment, a configuration variable ``websauna.site_url`` is passed as the base URL.

    See also :func:`make_dummy_request`.

    TODO: Split this to two different functions: one for existing dbsession and one for where dbsession is connected.

    :param dbsession: Use existing dbsession or set to ``None`` to generate a new dbsession and transaction manager. None that this TM is not the thread local transaction manager in ``transaction.mananger``.
    :param registry: Configuration registry
    :param path: Path being requested.
    :return: Current request.
    """
    base_url = registry.settings.get("websauna.site_url", None)
    # TODO: Honour request_factory here
    request = Request.blank(path, base_url=base_url)
    request.registry = registry
    request.user = None
    request.view_name = ''

    # This will create request.tm, others
    apply_request_extensions(request)

    if dbsession:
        # Use the provided dbsession for this request
        request.dbsession = dbsession
        if hasattr(dbsession, "transaction_manager"):
            request.tm = request.transaction_manager = dbsession.transaction_manager
    else:
        # Create a new dbsession and transaction manager for this request
        tm = TransactionManager()
        dbsession = create_dbsession(request.registry, tm)
        request.dbsession = dbsession
        request.tm = request.transaction_manager = tm

        def terminate_session(request):
            # Close db session at the end of the request and return the db connection back to the pool
            dbsession.close()

        request.add_finished_callback(terminate_session)

    return request



"""Authentication tweens."""
# Standard Library
import logging

# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.registry import Registry

# SQLAlchemy
from sqlalchemy import orm

# Websauna
from websauna.system.core import messages
from websauna.system.http import Request


try:
    from pyramid_tm import reify  # noQA
    good_reify = True
except ImportError:
    good_reify = False


logger = logging.getLogger(__name__)


class SessionInvalidationTweenFactory:
    """Tween to detect invalidated sessions and redirect user back to home.

    This tween checks if the current session is logged in user and there has been authentication sensitive changes to this user. In this case all user sessions should be logged out.
    """

    def __init__(self, handler, registry: Registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request: Request):
        user = request.user
        if user:
            try:
                session_authenticated_at = request.session.get("authenticated_at")

                # User was deauthenticatd in this request for some reason
                if session_authenticated_at:

                    if not user.is_valid_session(session_authenticated_at):
                        request.session.invalidate()
                        messages.add(request, kind="error", msg="Your have been logged out due to authentication changes.", msg_id="msg-session-invalidated")
                        logger.info("User log out forced due to security sensitive settings change, user %s, session id %s", user, request.session.session_id)
                        return HTTPFound(request.application_url)
            except orm.exc.DetachedInstanceError:
                if good_reify:
                    # pyramid_tm 2.0
                    raise

                # TODO: pyramid_tm 2.0 needed,
                # now temporary just kill user object instead of failing with an internal error, so that development server doesn't fail with CSS etc. resources
                request.user = None

        response = self.handler(request)
        return response



"""Celery process."""
# Standard Library
import logging
import os
import sys
import typing as t

# Pyramid
import plaster

from celery.loaders.base import BaseLoader
from celery.signals import setup_logging as _setup_logging_signal

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.cmdline import setup_logging
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.http.utils import make_routable_request
from websauna.system.model.retry import ensure_transactionless
from websauna.system.task.celery import parse_celery_config
from websauna.utils.secrets import read_ini_secrets


logger = logging.getLogger(__name__)


#: Passed through Celery loader mechanism
ini_file = ''


def usage_message(argv: t.List[str]):
    """Display usage message and exit.

    :param argv: Command line arguments.
    :raises sys.SystemExit:
    """
    cmd = os.path.basename(argv[0])
    msg = (
        'usage: {cmd} <config_uri> -- worker\n'
        '(example: "{cmd} ws://conf/production.ini -- worker")'
    ).format(cmd=cmd)
    feedback_and_exit(msg, status_code=1, display_border=False)


class WebsaunaLoader(BaseLoader):
    """Celery command line loader for Websauna.

    Support binding request object to Celery tasks and loading Celery settings through Pyramid INI configuration.
    """

    def get_celery_config(self, config_file: str) -> str:
        """Return celery configuration, from given config_file.

        :param config_file: Websauna configuration file.
        :return: Celery configuration, as a string.
        """
        value = None
        loader = plaster.get_loader(config_file)
        settings = loader.get_settings('app:main')
        secrets_file = settings.get('websauna.secrets_file')
        if secrets_file:
            secrets = read_ini_secrets(secrets_file)
            value = secrets.get('app:main.websauna.celery_config', '')
        value = value if value else settings.get('websauna.celery_config')
        if not value:
            raise RuntimeError('Could not find websauna.celery_config in {ini_file}'.format(ini_file=ini_file))
        return value

    def read_configuration(self, **kwargs) -> dict:
        """Load Celery config from Pyramid INI file.

        We need to be able to do this without ramping up full Websauna, because that's the order of the events Celery worker wants. This way we avoid circular dependencies during Celery worker start up.
        """
        value = self.get_celery_config(ini_file)
        config = parse_celery_config(value)
        return config

    def import_task_module(self, module):
        raise RuntimeError('Import Celery config directive is not supported. Use config.scan() to pick up tasks.')

    def register_tasks(self):
        """Inform Celery of all tasks registered through our Venusian-compatible task decorator."""
        # @task() decorator pick ups
        tasks = getattr(self.request.registry, "celery_tasks", [])

        for func, args, kwargs in tasks:
            decorator_args = [func] + list(args)
            self.app.task(*decorator_args, **kwargs)

    def on_worker_init(self):
        """This method is called when a child process starts."""
        # TODO Make sanity_check True by default,
        # but make it sure we can disable it in tests,
        # because otherwise tests won't run
        self.request = init_websauna(ini_file, sanity_check=False)

        #: Associate this process as Celery app object of the environment
        self.request.registry.celery = self.app

        #: Associate request for celery app, so
        #: task executor knows about Request object
        self.app.cmdline_request = self.request

        self.register_tasks()

    def on_task_init(self, task_id, task):
        """This method is called before a task is executed.

        Pass our request context to the task.

        http://docs.celeryproject.org/en/latest/userguide/tasks.html#context

        .. note ::

            The same request object is recycled over and over again. Pyramid does not have correctly mechanisms for having retryable request factory.

        """

        # TODO: How Celery handles retries?

        # We must not have on-going transaction when worker spawns a task
        # - otherwise it means init code has left transaction open
        ensure_transactionless("Thread local TX was ongoing when Celery fired up a new task {}: {}".format(task_id, task))

        # Each tasks gets a new request with its own transaction manager and dbsession
        request = make_routable_request(dbsession=None, registry=self.request.registry)

        task.request.update(request=request)


@_setup_logging_signal.connect
def fix_celery_logging(loglevel, logfile, format, colorize, **kwargs):
    """Fix Celery logging by re-enforcing our loggers after Celery messes up them."""
    setup_logging(ini_file)


def main(argv: t.List[str]=sys.argv):
    """Celery process entry point.

    Wrap celery command line script with our INI reader.

    .. note ::

        Make sure there is no global app = Celery() in any point of your code base,
        or this doesn't work.

    """
    global ini_file
    global request

    if len(argv) < 2:
        usage_message(argv)

    ini_file = get_config_uri(argv)
    if not ini_file.endswith(".ini"):
        sys.exit("The first argument must be a configuration file")

    if len(argv) >= 3:
        if not argv[2] == "--":
            raise RuntimeError("The second argument must be -- to signal command line argument passthrough")
        celery_args = sys.argv[3:]
    else:
        celery_args = []

    # https://github.com/celery/celery/issues/3405
    os.environ["CELERY_LOADER"] = "websauna.system.task.celeryloader.WebsaunaLoader"
    argv = ["celery"] + celery_args
    # Directly jump to Celery 4.0+ entry point
    from celery.bin.celery import main
    main(argv)



# Websauna
from websauna.system.admin.interfaces import IAdmin
from websauna.system.admin.modeladmin import ModelAdmin
from websauna.system.core.traversal import Resource


def get_admin(request) -> IAdmin:
    """Get hold of the default site admin interface root object."""
    admin_class = request.registry.queryUtility(IAdmin)
    return admin_class(request)


def get_admin_for_model(admin: IAdmin, model: type) -> Resource:
    """Return Admin resource for a model manager interface.

    """

    model_manager = admin["models"]

    if model.id not in model_manager:
        raise KeyError("No admin defined for model: {}".format(model))
    return model_manager[model.id]


def get_admin_resource_for_sqlalchemy_object(admin: IAdmin, instance: object) -> ModelAdmin.Resource:
    """Return ModelAdmin.Resource for an SQLAlchemy object.

    Example how to get an admin edit link for an SQLAlchemy object:

    .. code-block:: python

        resource = get_admin_resource_for_sqlalchemy_object(request.admin, asset)
        return request.resource_url(resource, "edit")


    :param admin: ``request.admin``
    :param instance: SQLAlchemy instance
    """
    admin = get_model_admin_for_sqlalchemy_object(admin, instance)
    res = admin.wrap_to_resource(instance)
    return res


def get_model_admin_for_sqlalchemy_object(admin: IAdmin, instance: object) -> ModelAdmin:
    """Return ModelAdmin resource for a SQLAlchemy object instance."""
    model_manager = admin["models"]

    registry = admin.request.registry
    model = instance.__class__

    model_admin_id = registry.model_admin_ids_by_model.get(model)
    assert model_admin_id, "No ModelAdmin configured for model {}".format(model)

    return model_manager[model_admin_id]


def get_admin_url_for_sqlalchemy_object(admin: IAdmin, instance: object, view_name: str=None) -> str:
    """Return direct URL to the admin view page of this object:

    Example:

    .. code-block:: python

        link = get_admin_url_for_sqlalchemy_object(request.admin, choice, view_name="edit")

    :param admin: Admin root object
    :param instance: SQLAlchemy object
    """
    model_admin = get_model_admin_for_sqlalchemy_object(admin, instance)
    return model_admin.get_object_url(instance, view_name=view_name)



"""Abstraction over generating Deform forms from different CRUD models.

This is used by CRUD/admin to create a form for a model to edit. Model doesn't need to be any particular kind. SQLAlchemy is supported by default.
"""
# Standard Library
import typing as t
from abc import abstractmethod

# Pyramid
import colander
import deform

# Websauna
from websauna.system.form.editmode import EditMode
from websauna.system.form.fieldmapper import DefaultSQLAlchemyFieldMapper
from websauna.system.form.resourceregistry import ResourceRegistry
from websauna.system.form.schema import add_csrf
from websauna.system.http import Request

from . import Resource


def default_schema_binder(schema: colander.Schema, request: Request, context: Resource, **kwargs) -> colander.Schema:
    """Initialize Colander field dynamic default values.

    By default, don't do anything and pass ``request`` and ``context`` to schema.

    :param schema: A colander Schema.
    :param request: Current HTTP Request.
    :param context: Traversal context
    :param kwargs: Optional keyword arguments (not used atm)
    :return: Bound colander.Schema
    """
    return schema.bind(request=request, context=context)


def _noop_customizer(**kwargs):
    return


class FormGenerator:
    """Generate a Deform form.

    :param schema_customizer: is a callback ``customize_schema(schema=schema, request=request, context=context, mode=mode, model=model)`` and can dynamically adjust generated form schema after its generation. Customizers edits schema in place and does not return anything.

    :param schema_binder: is a callback ``bind_schema(schema=schema, request=request, context=context, mode=mode, model=model)``. It should call ``schema.bind()`` and pass any extra arguments for Colander schema binding.

    By default we usually pass:

    .. code-block:: python

        schema.bind(request=request, context=context)

    """

    def __init__(self, schema_customizer: t.Callable=None, schema_binder: t.Callable=None):
        """Initialize Form Generator."""
        self.schema_customizer = schema_customizer
        self.schema_binder = schema_binder

    @abstractmethod
    def generate_form(self, request: Request, context: Resource, mode: EditMode, buttons: t.List[deform.Button], model: type) -> deform.Form:
        """Generate form.

        :param request: Current HTTP request
        :param context: Traversal context
        :param mode: Of of different form edit modes
        :param buttons: List of form buttons
        :param model: SQLAlchemy model class
        :return: Constructed form object
        """

    def create_deform(self, schema: colander.Schema, request: Request, context: Resource, mode: EditMode, buttons: t.List[deform.Button], model: type) -> deform.Form:
        """Create a Deform based on a given generated schema.

        This method does the following:

            * Adds CSRF token to schema

            * Calls schema customizer

            * Binds schema to request and context or calls custom schema binder

            * Creates the form instance

        :param schema: Schema to be used.
        :param request: Current HTTP request.
        :param context: Traversal context.
        :param mode: Form mode.
        :param buttons: List of buttons to be added.
        :param model: SQLAlchemy model class
        :return: Constructed form object
        """
        # Make sure we have CSRF token
        add_csrf(schema)

        # Customize schema
        customizer = self.schema_customizer or _noop_customizer
        customizer(schema=schema, request=request, context=context, mode=mode, model=model)

        # Bind schema
        binder = self.schema_binder or default_schema_binder
        schema = binder(schema=schema, request=request, context=context, mode=mode, model=model)

        form = deform.Form(schema, buttons=buttons, resource_registry=ResourceRegistry(request))
        return form


class SQLAlchemyFormGenerator(FormGenerator):
    """Automatically create a Deform form from underlying SQLAlchemy model.

    :param includes: This is a list of [column name, colander.SchemaNode] we extract from the SQLAlchemy model and generate fields for. If you supply a column name as a string the schema type and widget is automatically resolved. If you give ``colander.SchemaNode(name=name)`` then this schema type is explicitly used, regardless if such a column exist in the model or not.

    If ``includes`` is set to ``None`` it tries to convert all columns to fields it sees on a model.

    For example use case see :py:class:`websauna.system.user.adminviews.UserAdd`.
    """

    def __init__(self, includes=None, field_mapper=DefaultSQLAlchemyFieldMapper(), customize_schema: t.Callable=None, schema_binder: t.Callable=None):
        """Initialize SQLAlchemyFormGenerator."""
        self.includes = includes
        self.field_mapper = field_mapper
        super(SQLAlchemyFormGenerator, self).__init__(customize_schema, schema_binder)

    def generate_form(self, request: Request, context: Resource, mode: EditMode, buttons: t.List[deform.Button], model: type) -> deform.Form:
        """Generate form.

        :param request: Current HTTP request
        :param context: Traversal context
        :param mode: Of of different form edit modes
        :param buttons: List of form buttons
        :param model: SQLAlchemy model class
        :return: Constructed form object
        """
        schema = self.field_mapper.map(mode, request, context, model, self.includes, nested=True)
        return self.create_deform(schema, request, context, mode, buttons, model)



# Standard Library
import os

# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config

from pyramid_notebook import startup
from pyramid_notebook.views import notebook_proxy as _notebook_proxy
from pyramid_notebook.views import shutdown_notebook as _shutdown_notebook
from pyramid_notebook.views import launch_notebook

# Websauna
from websauna.system.model.meta import Base


# Don't do changedir as it doesn't work. TODO: fix bad change_dir in upstream
WEBSAUNA_BOOSTRAP = """
from pkg_resources import load_entry_point
from pyramid_notebook.utils import change_directory

# We need to use this to trigger a proper namespaced package loading
# due to different approaches between pip / easy_install / python setup.py develop
# (it doesn't matter which entry point we load as long as it's from websauna package)
entry_point  = load_entry_point('websauna', 'console_scripts', 'ws-shell')

from websauna.system.devop.cmdline import init_websauna_script_env

# Our development.ini, production.ini, etc.
config_file = '{config_uri}'

with change_directory('{cwd}'):
    script_env = init_websauna_script_env(config_file)
    globals().update(script_env)

"""


#: Include our database session in notebook so it is easy to query stuff right away from the prompt
SCRIPT = """
dbsession = request.dbsession
from websauna.utils.time import now
import sqlalchemy
"""


GREETING = """
* **sqlalchemy** - sqlachemy module
* **dbsession** - SQLAlchemy database session
* **now()** - UTC time as timezone aware datetime object
* **initializer** - websauna.system.Initializer instance
"""


def launch_context_sensitive_shell(request, extra_script="", extra_greeting=""):
    """Launch a IPython Notebook.

    :param extra_script: Extra script executed on the launch of this notebook
    :param extra_greeting: Extra text in the greeting Markdown for this launch
    """
    nb = {}

    # Pass around the Pyramid configuration we used to start this application
    global_config = request.registry.settings["websauna.global_config"]

    # Get the reference to our Pyramid app config file and generate Notebook
    # bootstrap startup.py script for this application
    config_file = global_config["__file__"]
    startup.make_startup(nb, config_file, bootstrap_py=WEBSAUNA_BOOSTRAP, cwd=os.getcwd())
    startup.add_script(nb, SCRIPT)
    startup.add_greeting(nb, GREETING)

    #: Include all our SQLAlchemy models in the notebook variables
    startup.include_sqlalchemy_models(nb, Base)

    startup.add_script(nb, extra_script)
    startup.add_greeting(nb, extra_greeting)

    return launch_notebook(request, request.user.username, notebook_context=nb)


# TODO: Broken at the moment because CSRF conflict
@view_config(route_name="notebook_proxy", permission="shell", require_csrf=False)
def notebook_proxy(request):
    """Proxy IPython Notebook requests to the upstream server.

    A special ``shell`` permission is needed to access this view. See :ref:`websauna.superusers`.
    """
    return _notebook_proxy(request, request.user.username)


@view_config(route_name="admin_shell", permission="shell")
def admin_shell(request):
    """Open admin shell with default parameters for the user.

    A special ``shell`` permission is needed to access this view. See :ref:`websauna.superusers`.
    """
    # Make sure we have a logged in user
    return launch_context_sensitive_shell(request)


@view_config(route_name="shutdown_notebook", permission="shell")
def shutdown_notebook(request):
    """Shutdown the notebook of the current user."""
    _shutdown_notebook(request, request.user.username)
    return HTTPFound(request.route_url("home"))



"""HTTP 403 Forbidden view."""

# Pyramid
import transaction
from pyramid.renderers import render
from pyramid.response import Response
from pyramid.view import forbidden_view_config


@forbidden_view_config()
def forbidden(request):
    # The template rendering opens a new transaction which is not rolled back by Pyramid transaction machinery, because we are in a very special view. This tranaction will cause the tests to hang as the open transaction blocks Base.drop_all() in PostgreSQL. Here we have careful instructions to roll back any pending transaction by hand.
    html = render('core/forbidden.html', {}, request=request)
    resp = Response(html)
    resp.status_code = 403

    # Hint pyramid_redis_session not to generate any session cookies for this response
    resp.cache_control.public = True

    transaction.abort()
    return resp



# Pyramid
from pyramid.view import view_config

# Websauna
from websauna.system.admin.modeladmin import ModelAdmin


class Shell:
    """Notebook shell opener.

    Prepolate shell with this object through dbsession query.
    """

    def __init__(self, context, request):
        self.request = request
        self.context = context

    @view_config(context=ModelAdmin.Resource, name="shell", route_name="admin", permission='shell')
    def shell(self):

        from websauna.system.notebook.views import launch_context_sensitive_shell

        obj = self.context.get_object()
        extra_script = "obj = dbsession.query({}).get({})".format(obj.__class__.__name__, obj.id)
        extra_greeting = "* **obj** {}".format(self.context.get_title())
        return launch_context_sensitive_shell(self.request, extra_script, extra_greeting)



# Standard Library
import ipaddress
from datetime import timedelta

# Pyramid
import deform
import transaction
from deform import Button

from webtest import TestApp

# Websauna
import websauna
from websauna.system.tests.utils import EMAIL
from websauna.system.tests.utils import PASSWORD
from websauna.system.tests.utils import create_user
from websauna.utils.time import now


def get_user(dbsession):
    from websauna.system.user.models import User
    return dbsession.query(User).get(1)


def test_login(web_server, browser, dbsession, init):
    """Login an user."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit(web_server)

    b.click_link_by_text("Sign in")

    assert b.is_element_present_by_css("#login-form")

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    # After login we see a profile link to our profile
    assert b.is_element_present_by_css("#nav-logout")


def test_logout(web_server, browser, dbsession, init):
    """Log out."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit("{}/{}".format(web_server, "login"))

    assert b.is_element_present_by_css("#login-form")

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    assert b.is_element_present_by_css("#msg-you-are-logged-in")
    b.find_by_css("#nav-logout").click()

    # Anonynous again
    assert b.is_element_present_by_css("#msg-logged-out")
    assert not b.is_element_present_by_css("#nav-logout")

    # We should see the log in form
    assert b.is_element_present_by_css("#login-form")


def test_last_login_ip(web_server, browser, dbsession, init):
    """Record last log in IP correctly."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    with transaction.manager:
        user = get_user(dbsession)
        assert not user.last_login_ip

    b = browser
    b.visit(web_server)

    b.click_link_by_text("Sign in")

    assert b.is_element_present_by_css("#login-form")

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    assert b.is_element_present_by_css("#msg-you-are-logged-in")

    with transaction.manager:
        user = get_user(dbsession)
        assert user.last_login_ip in [ipaddress.IPv4Address("127.0.0.1"), ipaddress.IPv6Address("::1")]


def test_forget_password(web_server, browser, dbsession, init):
    """Reset password by email."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit(web_server)

    b.click_link_by_text("Sign in")

    assert b.is_element_present_by_css("#login-form")

    b.click_link_by_text("Forgot your password?")
    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", EMAIL)
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-check-email")

    with transaction.manager:
        user = get_user(dbsession)
        activation_code = user.activation.code

    b.visit("{}/reset-password/{}".format(web_server, activation_code))
    assert b.is_element_present_by_css("#reset-password-form")

    # Friendly name should be visible
    assert b.is_text_present("example@example.com")
    b.fill("password", "yyy")
    b.fill("password-confirm", "yyy")
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-password-reset-complete")

    b.fill("username", EMAIL)
    b.fill("password", "yyy")
    b.find_by_name("login_email").click()

    assert b.is_element_present_by_css("#nav-logout")


def test_forget_password_bad_user(web_server, browser, dbsession, init):
    """Reset password by email."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit(web_server + "/login")

    assert b.is_element_present_by_css("#login-form")

    b.click_link_by_text("Forgot your password?")
    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", "foo@example.com")
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css(".error-msg-detail")


def test_forget_password_disabled_user(web_server, browser, dbsession, init):
    """Reset password by email."""

    with transaction.manager:
        u = create_user(dbsession, init.config.registry)
        u.enabled = False

    b = browser
    b.visit(web_server + "/login")

    assert b.is_element_present_by_css("#login-form")

    b.click_link_by_text("Forgot your password?")
    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", EMAIL)
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-cannot-reset-password")


def test_bad_forget_password_activation_code(web_server, browser, dbsession):
    """Reset password by email."""
    b = browser
    b.visit("{}/reset-password/xxxx".format(web_server))

    # Check we get the pimped up not found page
    assert b.is_element_present_by_css("#not-found")


def test_login_forget_password_email_send(web_server, browser, dbsession, init):
    """Send out the reset password by email, but do not answer to it, instead directly login."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit(web_server)

    b.find_by_css("#nav-sign-in").click()

    assert b.is_element_present_by_css("#login-form")

    b.click_link_by_text("Forgot your password?")
    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", EMAIL)
    b.find_by_name("submit").click()

    b.visit("{}/login".format(web_server))

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()
    assert b.is_element_present_by_css("#msg-you-are-logged-in")


def test_forget_password_expired_token(web_server, browser, dbsession, init):
    """Reset password by email."""

    with transaction.manager:
        create_user(dbsession, init.config.registry)

    b = browser
    b.visit(web_server + "/forgot-password")

    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", EMAIL)
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-check-email")

    with transaction.manager:
        user = get_user(dbsession)
        activation = user.activation
        activation.expires_at = now() - timedelta(days=365)
        activation_code = activation.code

    b.visit("{}/reset-password/{}".format(web_server, activation_code))
    assert b.is_element_present_by_css("#not-found")


def test_customize_login(paster_config):
    """Customizing login form works."""

    class CustomLoginForm(deform.Form):

        def __init__(self, *args, **kwargs):
            login_button = Button(name="login_email", title="Login by fingerprint", css_class="btn-lg btn-block")
            kwargs['buttons'] = (login_button,)
            super().__init__(*args, **kwargs)

    class Initializer(websauna.system.DemoInitializer):

        def configure_user_forms(self):

            from websauna.system.user import interfaces

            # This will set up all default forms as shown in websauna.system.Initializer.configure_user_forms
            super(Initializer, self).configure_user_forms()

            # Override the default login form with custom one
            unregister_success = self.config.registry.unregisterUtility(provided=interfaces.ILoginForm)
            assert unregister_success, "No default form register"
            self.config.registry.registerUtility(CustomLoginForm, interfaces.ILoginForm)

    global_config, app_settings = paster_config
    init = Initializer(global_config, app_settings)
    init.run()
    app = TestApp(init.make_wsgi_app())
    resp = app.get("/login")

    assert "Login by fingerprint" in resp.text



"""See that beat runs scheduled tasks correctly."""

# Standard Library
import os
import subprocess
import time

from flaky import flaky

# Websauna
from websauna.system.core.redis import get_redis


def run_worker_and_beat(ini_file):

    cmdline = "ws-celery ws://{} -- worker --loglevel=debug".format(ini_file)

    # You can start manually ws-celery websauna/tests/task-test.ini -- worker --loglevel=debug
    # and set worker = False
    worker = None

    if worker is None:
        worker = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        time.sleep(2.0)

        worker.poll()
        if worker.returncode is not None:
            print(worker.stdout.read().decode("utf-8"))
            print(worker.stderr.read().decode("utf-8"))
            raise AssertionError("Celery worker process did not start up: {}".format(cmdline))

    # You can run manually ws-celery beat -A websauna.system.task.celery.celery_app --ini websauna/tests/scheduler-test.ini
    cmdline = "ws-celery {} -- beat".format(ini_file)
    beat = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    time.sleep(1.0)
    beat.poll()
    if beat.returncode is not None:
        if worker:
            worker.terminate()
            AssertionError("Beat process did not start up")

    return worker, beat


@flaky
def test_celery_beat(init):
    """Scheduled tasks run properly on the celery worker + celery beat process."""

    ini_file = os.path.join(os.path.dirname(__file__), "task-test.ini")
    worker, beat = run_worker_and_beat(ini_file)

    try:
        # Reset test database
        redis = get_redis(init.config.registry)
        redis.delete("foo", "bar")

        foo = "no read"
        deadline = time.time() + 10
        while time.time() < deadline:
            redis = get_redis(init.config.registry)
            # scheduledtasks.ticker should beat every second and reset values in Redis
            foo = redis.get("foo")
            if foo:
                break
            time.sleep(0.5)

        if worker:
            assert worker.returncode is None

        if beat:
            assert beat.returncode is None

        if foo != b"xoo":
            # TravisCI headless debugging
            # print(worker.stdout.read().decode("utf-8"))
            # print(worker.stderr.read().decode("utf-8"))
            # print(beat.stdout.read().decode("utf-8"))
            # print(beat.stderr.read().decode("utf-8"))
            pass

        assert foo == b"xoo"  # Set back by its original value by 1 second beat

    finally:
        try:
            if worker:
                worker.terminate()
        except ProcessLookupError:
            pass

        try:
            beat and beat.terminate()
        except ProcessLookupError:
            pass



import requests


SPOOF_CSRF_JS = """
(function() {
console.log('Preparing CSRF spoofing');
$("input[name='csrf_token']").val("xxx");
console.log("Finished geospoofing");
})();
"""


def test_internal_server_error(customized_web_server, browser):
    """When things go KABOOM show a friendly error page."""

    web_server = customized_web_server({"websauna.log_internal_server_error": False})

    b = browser
    b.visit("{}/error-trigger".format(web_server))

    # After login we see a profile link to our profile
    assert b.is_element_visible_by_css("#it-is-broken")


def test_not_found(web_server, browser):
    """Show not found page on unknown URL."""

    b = browser
    missing_url = "{}/foobar".format(web_server)
    b.visit(missing_url)

    assert b.is_element_visible_by_css("#not-found")

    # Splinter b.status_code is not reliable
    resp = requests.get(missing_url)
    assert resp.status_code == 404


def test_forbidden(web_server, browser):
    """Access controlled page should not be available."""

    b = browser
    b.visit("{}/admin/".format(web_server))

    assert b.is_element_visible_by_css("#forbidden")


def test_csrf_fail(web_server, browser):
    """Bad CSRF token shoud give us an error page."""

    b = browser
    b.visit(web_server + "/login")
    b.fill("username", "x")
    b.fill("password", "y")

    b.evaluate_script(SPOOF_CSRF_JS.replace("\n", " "))

    b.find_by_name("login_email").click()

    # Only present on our custom page
    assert b.is_element_present_by_css("#heading-bad-csrf-token")



# Standard Library
from datetime import timedelta

# Pyramid
import transaction

# Websauna
from websauna.system.user.models import Activation
from websauna.system.user.models import User
from websauna.utils.time import now


EMAIL = "example@example.com"
PASSWORD = "ToholamppiMadCowz585"


def get_user(dbsession):
    from websauna.system.user.models import User
    return dbsession.query(User).get(1)


def test_register_email(web_server, browser, dbsession):
    """Register on the site and login after activation."""

    b = browser
    b.visit(web_server)

    b.find_by_css("#nav-sign-up").click()

    assert b.is_element_present_by_css("#sign-up-form")

    b.fill("email", EMAIL)
    b.fill("password", PASSWORD)
    b.fill("password-confirm", PASSWORD)

    b.find_by_name("sign_up").click()

    assert b.is_element_present_by_css("#waiting-for-activation")

    # Now peek the Activation link from the database
    user = get_user(dbsession)
    assert user.activation.code

    activation_link = "{}/activate/{}".format(web_server, user.activation.code)

    b.visit(activation_link)

    assert b.is_element_present_by_css("#sign-up-complete")

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    # After login we see a profile link to our profile
    assert b.is_element_present_by_css("#nav-logout")


def test_register_email_activation_expired(web_server, browser, dbsession):
    """Register on the site, let the first activation expire, register through reset password."""

    b = browser

    b.visit(web_server + "/register")
    assert b.is_element_visible_by_css("#sign-up-form")

    b.fill("email", EMAIL)
    b.fill("password", PASSWORD)
    b.fill("password-confirm", PASSWORD)

    b.find_by_name("sign_up").click()

    assert b.is_element_visible_by_css("#waiting-for-activation")

    with transaction.manager:
        a = dbsession.query(Activation).get(1)
        u = dbsession.query(User).get(1)  # noQA
        a.expires_at = now() - timedelta(days=365)
        activation_code = a.code

    activation_link = "{}/activate/{}".format(web_server, activation_code)
    b.visit(activation_link)

    assert b.is_element_present_by_css("#not-found")

    b.visit(web_server + "/login")
    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    # Cannot login because activation expired
    assert b.is_element_present_by_css("#msg-authentication-failure")

    # Reset password
    b.visit(web_server + "/forgot-password")
    assert b.is_element_present_by_css("#forgot-password-form")
    b.fill("email", EMAIL)
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-check-email")

    with transaction.manager:
        user = get_user(dbsession)
        activation = user.activation
        activation_code = activation.code

    b.visit("{}/reset-password/{}".format(web_server, activation_code))

    b.fill("password", "yyy")
    b.fill("password-confirm", "yyy")
    b.find_by_name("submit").click()

    assert b.is_element_present_by_css("#msg-password-reset-complete")

    b.fill("username", EMAIL)
    b.fill("password", "yyy")
    b.find_by_name("login_email").click()

    assert b.is_element_present_by_css("#msg-you-are-logged-in")



"""The dreaded 500 page."""
# Standard Library
import logging

# Pyramid
from pyramid.renderers import render
from pyramid.response import Response
from pyramid.settings import asbool
from pyramid.view import view_config

# Websauna
from websauna.system.core.events import InternalServerError


try:
    from pyramid_tm.reify import can_access_transaction_in_excview
    from pyramid_tm.reify import reset_transaction_aware_properties
    HAS_NEW_PYRAMID_TM = True
except ImportError:
    HAS_NEW_PYRAMID_TM = False


logger = logging.getLogger(__name__)


@view_config(context=Exception, require_csrf=False)
def internal_server_error(context, request):
    """Generate the default internal server error page when exception falls through from a view.

    This view is marked as CSRF exempt, so that HTTP POST requests to API endpoints do not cause additional Bad CSRF exception when HTTP 500 Internal server error is raised.

    Also see https://github.com/Pylons/pyramid_tm/issues/40
    """

    if HAS_NEW_PYRAMID_TM:
        if not can_access_transaction_in_excview(context, request):
            # Kill some of db aware properties in the case templates
            # might accidentally touch them
            request.__dict__["user"] = None

        else:
            # We should have db and request.user available,
            # we have just started a new transction for this view
            reset_transaction_aware_properties(request)
    else:
        request.__dict__["user"] = None

    # Tell Sentry handler to log this exception on sentry
    request.registry.notify(InternalServerError(context, request))

    if asbool(request.registry.settings.get("websauna.log_internal_server_error", True)):
        logger.exception(context)

    html = render('core/internalservererror.html', {}, request=request)
    resp = Response(html)
    resp.status_code = 500

    # Hint pyramid_redis_session not to generate any session cookies for this response
    resp.cache_control.public = True

    # Make sure nothing is written or no transaction left open on 500
    request.tm.abort()

    return resp



"""Resolve principals (groups and pseudo roles) for ACL."""
# Standard Library
import typing as t

# Pyramid
from pyramid.security import Authenticated
from pyramid.settings import asbool
from pyramid.settings import aslist

# Websauna
from websauna.system.http import Request
from websauna.system.user.utils import get_user_registry


def resolve_principals(session_token: str, request: Request) -> t.Optional[t.List[str]]:
    """Get applied groups and other for the user.

    This is a callback for :py:class:`pyramid.authentication.SessionAuthenticationPolicy`.

    * List all groups as ``group:admin`` style strings

    * List super user as ``superuser:superuser`` style string

    :return: None if the user is not logged in, otherwise list of principals assigned to the user site wide.
    """

    user_registry = get_user_registry(request)
    user = user_registry.get_user_by_session_token(session_token)
    if not user:
        return None

    settings = request.registry.settings

    # Read superuser names from the config
    superusers = aslist(settings.get("websauna.superusers"))

    admin_as_superuser = asbool(settings.get("websauna.admin_as_superuser", False))

    if user_registry.can_login(user):

        # Users always get Authenticated special principal
        principals = [Authenticated]

        # All groups for this user
        principals += ['group:{}'.format(g.name) for g in user_registry.get_groups(user)]

        # Give principal for per user permissions
        principals += ['user:{}'.format(user.id)]

        # Allow superuser permission
        if user.username in superusers or user.email in superusers:
            # Superuser explicitly listed in the configuration
            principals.append("superuser:superuser")
        elif admin_as_superuser and ("group:admin" in principals):
            # Automatically promote admins to superusers when doing local development
            principals.append("superuser:superuser")

        return principals

    # User not found, user disabled
    return None



# Pyramid
from pyramid.registry import Registry
from pyramid.util import DottedNameResolver

from pyramid_mailer import IMailer


def get_mailer(registry) -> IMailer:
    """Get the active mailer.
    :param registry:
    :return: IMailer
    """
    return registry.getUtility(IMailer)


def create_mailer(registry: Registry) -> IMailer:
    """Create a new mailer instance.

    """

    settings = registry.settings

    # Empty values are not handled gracefully, so mutate them here before passing forward to mailer
    if settings.get("mail.username", "x") == "":
        settings["mail.username"] = None

    if settings.get("mail.password", "x") == "":
        settings["mail.password"] = None

    mailer_class = settings.get("websauna.mailer", "")
    if mailer_class in ("mail", ""):
        # TODO: Make mailer_class explicit so we can dynamically load pyramid_mail.Mailer
        # Default
        from pyramid_mailer import mailer_factory_from_settings
        mailer = mailer_factory_from_settings(settings)
    else:
        # debug backend
        resolver = DottedNameResolver()
        mailer_cls = resolver.resolve(mailer_class)
        mailer = mailer_cls()

    return mailer



"""Extra logging context."""
# Standard Library
import logging

# Pyramid
from pyramid.request import Request

# SQLAlchemy
import sqlalchemy


logger = logging.getLogger(__name__)


def is_good_sqlalchemy_object(obj):
    """Let's not cause exceptions in exception handler/logger."""
    state = sqlalchemy.inspect(obj)
    return not state.detached


def get_logging_user_context(request: Request) -> dict:
    """Capture some extra user-specific information from the logging context.

    :return: Dict containing human readable user parameters to help identify the user on this request
    """
    try:
        user = getattr(request, "user", None)
    except sqlalchemy.exc.InvalidRequestError:
        # We had a rollback and could not capture the user,
        # because user has been rolled back
        user = None

    user_context = {}
    try:
        if user:
            if is_good_sqlalchemy_object(user):
                # Add additional user context to the logged exception
                username = getattr(user, "friendly_name", None) or getattr(user, "username", None) or str(user)
                email = getattr(user, "email", None)
                user_context.update(dict(user=username, email=email))
            else:
                user_context.update(dict(detached=True))

        # All the session data as JSON
        session = getattr(request, "session", None)
        if session:
            session = dict(session.items())
            user_context.update(dict(session=session))
        else:
            user_context.update(dict(session="No session data available in internal_server_error()"))

        user_context["ip"] = request.client_addr

        # TODO: Make this more generic
        # https://support.cloudflare.com/hc/en-us/articles/200168236-What-does-CloudFlare-IP-Geolocation-do-
        user_context["cloudflare_country"] = request.headers.get("cf-ipcountry")

        return user_context
    except Exception as e:
        logger.exception(e)
        logger.error("Failed to capture user context %s", request)
        return {}



"""Sending out HTML and plain text email."""
# Standard Library
import logging
import typing as t
from email.header import Header
from email.utils import formataddr

# Pyramid
from pyramid.renderers import render
from pyramid.settings import asbool
from transaction import TransactionManager

import premailer
from pyramid_mailer.message import Message

# Websauna
from websauna.system.http import Request
from websauna.system.mail.utils import create_mailer


logger = logging.getLogger(__name__)


def render_templated_mail(request: Request, template: str, context: dict) -> t.Tuple[str, str, str]:
    """Render an email that is divided to three template files.

    The email is assembled from three different templates:

    * Read subject from a subject specific template $template.subject.txt

    * Generate HTML email from HTML template, $template.body.html

    * Generate plain text email from HTML template, $template.body.txt

    Make sure you have configured your template engine (Jinja 2) to read TXT templates beside HTML.

    HTML body goes through Premailer transform step, inlining any CSS styles.

    :param template: Template filename base string for template tripled (subject, HTML body, plain text body). For example ``email/my_message`` would map to templates ``email/my_message.subject.txt``, ``email/my_message.body.txt``, ``email/my_message.body.html``

    :param context: Template context dictionary passed to the rendererer

    :return: Tuple(subject, text_body, html_body)
    """

    subject = render(template + ".subject.txt", context, request=request)
    subject = subject.strip()

    html_body = render(template + ".body.html", context, request=request)
    text_body = render(template + ".body.txt", context, request=request)

    # Inline CSS styles
    html_body = premailer.transform(html_body)

    return subject, text_body, html_body


def send_templated_mail(request: Request, recipients: t.List, template: str, context: dict, sender=None, immediate=None, tm: t.Optional[TransactionManager]=None) -> t.Tuple[str, str, str]:
    """Send out templatized HTML and plain text emails.

    Each HTML email should have a plain text fallback. Premailer package is used to convert any CSS styles in HTML email messages to inline, so that email clients display them.

    The email is assembled from three different templates:

    * Read subject from a subject specific template $template.subject.txt

    * Generate HTML email from HTML template, $template.body.html

    * Generate plain text email from HTML template, $template.body.txt

    Make sure you have configured your template engine (Jinja 2) to read TXT templates beside HTML.

    Example to send email with a custom From header, using ``myapp/email/welcome.body.html`` as the body template:

    .. code-block:: python

        from email.header import Header
        from email.utils import formataddr

        # I never receive any letters
        to_email = "mikko@example.com"

        # Build a proper From header with both name and email,
        # otherwise the sender defaults to the site INI settings
        sender = formataddr((str(Header(order.get_sender_name(), 'utf-8')), order.get_from_email()))

        # Pass some variables to subject and body templates
        context = dict(greeting="Welcome human!", tag_line="We come in peace")

        send_templated_mail(self.request, [to_email], "myapp/email/welcome", context=context, sender=sender)


    :param request: HTTP request, passed to the template engine. Request configuration is used to get hold of the configured mailer.

    :param recipients: List of recipient emails

    :param template: Template filename base string for template tripled (subject, HTML body, plain text body). For example ``email/my_message`` would map to templates ``email/my_message.subject.txt``, ``email/my_message.body.txt``, ``email/my_message.body.html``

    :param context: Template context dictionary passed to the rendererer

    :param sender: Override the sender email - if not specific use the default set in the config as ``mail.default_sender``

    :param immediate: Set True to send to the email immediately and do not wait the transaction to commit. This is very useful for debugging outgoing email issues in an interactive traceback inspector. If this is ``None`` then use setting ``mail.immediate`` that defaults to ``False``.

    :param tm: Give transaction manager that is used instead of ``request.tm`` for the commit hook.

    :return: tuple(subject, text_body, html_body)
    """

    assert recipients
    assert len(recipients) > 0
    assert type(recipients) != str, "Please give a list of recipients, not a string"

    for r in recipients:
        assert r, "Received empty recipient when sending out email {}".format(template)

    subject, text_body, html_body = render_templated_mail(request, template, context)

    # Have some logging by default, as inspecting mail issues is gruesome devops tasks
    logger.info("Sending out email to:%s subject:%s", recipients, subject)

    if not sender:
        sender = request.registry.settings["mail.default_sender"]

        # Add enveloped From:
        sender_name = request.registry.settings.get("mail.default_sender_name")
        if sender_name:
            sender = formataddr((str(Header(sender_name, 'utf-8')), sender))

    message = Message(subject=subject, sender=sender, recipients=recipients, body=text_body, html=html_body)
    message.validate()

    if not tm:
        tm = getattr(request, "tm", None)

    # TODO: Fix upstream pyramid_mailer

    # mailer = get_mailer(request)

    # We need to reconstruct mailer instance every time because it assumes thread local TM.
    # and we want to rewrite this value for every request
    mailer = create_mailer(request.registry)

    # Don't use the default ThreadLocal transaction manager as that won't fly with Celery.
    # Note we do this only for SMTP services, not for other more archaid mail deliveries.
    if tm:
        if hasattr(mailer, "direct_delivery"):
            # Not needed for dummy mailer
            mailer.direct_delivery.transaction_manager = tm

    if immediate is None:
        immediate = asbool(request.registry.settings.get("mail.immediate", False))

    if not tm:
        logger.warn("Warning: Implicit immediate email, because no transaction manager: %s", subject)
        immediate = True

    if immediate:
        mailer.send_immediately(message)
    else:
        mailer.send(message)

    return subject, text_body, html_body



"""Deform specific initialization logic."""
# Pyramid
import deform

from pkg_resources import resource_filename


def configure_zpt_renderer(search_path=()):
    """Initialize ZPT widget rendering for Deform forms.

    See https://github.com/Pylons/pyramid_deform/blob/master/pyramid_deform/__init__.py#L437
    """

    default_paths = deform.form.Form.default_renderer.loader.search_path
    paths = []
    for path in search_path:
        pkg, resource_name = path.split(':')
        paths.append(resource_filename(pkg, resource_name))
    deform.form.Form.default_renderer = deform.ZPTRendererFactory(
        tuple(paths) + default_paths)



"""Rolling time window counter and rate limit using Redis.

Use Redis sorted sets to do a rolling time window counters and limiters. These are useful for preventing denial of service, flood and reputation attack against site elements which trigegr outgoing action (email, SMS).

Example how to do a Colander validator which checks that the form has not been submitted too many times within the time period::

    import colander as s

    @c.deferred
    def throttle_invites_validator(node, kw):
        "Protect invite functionality from flood attacks."
        request = kw["request"]

        limit = int(request.registry.settings.get("trees.invite_limit", 60))

        def inner(node, value):
            # Check we don't have many invites going out
            if rollingwindow.check(request.registry, "invite_friends", window=3600, limit=limit):

                # Alert devops through Sentry
                logger.warn("Excessive invite traffic")

                # Tell users slow down
                raise c.Invalid(node, 'Too many outgoing invites at the moment. Please try again later.')

        return inner

Then you construct form::

    schema = schemas.InviteFriends(validator=schemas.throttle_invites_validator).bind(request=request)
    form = deform.Form(schema)

You can also exercise this code in tests::

    def test_flood_invite(web_server, browser, dbsession, init):
        "Overload invites and see we get an error message."

        b = browser
        with transaction.manager:
            create_user()

        # Set flood limit to two attempts
        init.config.registry.settings["trees.invite_limit"] = "2"

        # Clear Redis counter for outgoing invitations
        redis = get_redis(init.config.registry)
        redis.delete("invite_friends")

        # Login
        b.visit(web_server + "/login")
        b.fill("username", EMAIL)
        b.fill("password", PASSWORD)
        b.find_by_name("Log_in").click()

        def flood():
            b.visit("{}/invite-friends".format(web_server))
            b.find_by_css("#nav-invite-friends").click()
            b.fill("phone_number", "555 123 1234")
            b.find_by_name("invite").click()

        flood()
        assert b.is_text_present("Invite SMS sent")
        flood()
        assert b.is_text_present("Invite SMS sent")
        flood()
        assert b.is_text_present("Too many outgoing invites at the moment")


More info

* http://opensourcehacker.com/2014/07/09/rolling-time-window-counters-with-redis-and-mitigating-botnet-driven-login-attacks/

* http://redis.io/commands/zadd

"""

# Standard Library
import time

# Websauna
from websauna.system.core.redis import get_redis


def _check(redis, key, window=60, limit=50):

    # Expire old keys (hits)
    expires = time.time() - window
    redis.zremrangebyscore(key, '-inf', expires)

    # Add a hit on the very moment
    now = time.time()
    redis.zadd(key, now, now)

    # If we currently have more keys than limit,
    # then limit the action
    if redis.zcard(key) > limit:
        return True

    return False


def _get(redis, key):
    """ Get the current hits per rolling time window.

    :param redis: Redis client

    :param key: Redis key name we use to keep counter

    :return: int, how many hits we have within the current rolling time window
    """
    return redis.zcard(key)


def check(registry, key, window=60, limit=10):
    """Do a rolling time window counter hit.

    Use ``key`` to store the current hit rate in Redis.

    :param registry: Pyramid registry e.g. request.registry
    :param key: Redis key name we use to keep counter
    :param window: Rolling time window in seconds. Default 60 seconds.
    :param limit: Allowed operations per time window. Default 10 hits.

    :return: True is the maximum limit has been reached for the current time window
    """
    redis = get_redis(registry)
    return _check(redis, key, window, limit)


def get(registry, key):
    """Get the current hits per rolling time window.

     Use ``key`` to store the current hit rate in Redis.

    :param registry: Pyramid registry e.g. request.registry
    :param key: Redis key name we use to keep counter
    :return: int, how many hits we have within the current rolling time window
    """
    redis = get_redis(registry)
    return _check(redis, key)



"""Rendering helpers."""
# Websauna
from websauna.utils.orderedset import OrderedSet


class OnDemandResourceRenderer:
    """A JS and CSS resource registry which helps to allow widgets and such add JS and CSS files dynamically on the page.

    Usually instance of this class is available as ``request.on_demand_resource_renderer``.

    The best practice is to place all JavaScript files before the closing ``</body>`` tag, so that they do not block the the page rendering and decrease the page response time. However, out there exist a lot of badly written JavaScript code which does not honour this - they have ``<script>`` tags in the middle of ``<body>`` and even worse assume you have loaded dependencies, like jQuery, in the ``<head>``. Namely, Deform 2 form library does this.

    ``BodyRelocatableResourceRegistry`` contains the logic to determine if ``<script>`` tags should go to the end of the page (preferred) or to ``<head>`` (in case some javascript makes assumptions of this). If any widget requires resources assume JS must go to head.

    See also :py:attr`deform.widget.default_resources` which includes the list of default resources activated for any form on the page.

    More information

    * https://developers.google.com/speed/docs/insights/BlockingJS

    * https://developers.google.com/web/fundamentals/performance/critical-rendering-path/adding-interactivity-with-javascript#parser-blocking-vs-asynchronous-javascript
    """

    def __init__(self):
        self.resources = {
            "js": OrderedSet(),
            "css": OrderedSet(),
        }
        self.js_requires_head = False

    def request_resource(self, kind: str, resource_url: str, js_requires_head: bool=False):
        """A widget or something wants to place a CSS or JS file on the page rendering.

        :param kind: "js" or "css"
        :param resource_path: Resolved full URL to this resource
        :param js_requires_head: Move all JavaScript to <head> instead of </body> end. I.e. you have <script> tags in the middle of HTML.
        """
        self.resources[kind].add(resource_url)

        # If one JS wants to go head then everybody goes
        self.js_requires_head |= js_requires_head

    def is_js_in_head(self, request) -> bool:
        return self.js_requires_head

    def get_resources(self, kind) -> OrderedSet:
        """Get list of resource URLs to render.

        :param kind: "js" or "css"
        """
        return self.resources[kind]


def get_on_demand_resource_renderer(request):
    """Reify method for configuration."""
    return OnDemandResourceRenderer()



"""The default password hashing implementation using Argon 2.

Argon 2 is the winner algorithm of Password Hashing Competition 2012-2015.

More information

* https://github.com/hynek/argon2_cffi

* https://password-hashing.net/
"""
# Pyramid
from zope.interface import implementer

import argon2

# Websauna
from websauna.system.user.interfaces import IPasswordHasher


@implementer(IPasswordHasher)
class Argon2Hasher:
    """The default password hashing implementation using Argon 2."""

    def __init__(self):
        """Initialize Argon2Hasher."""
        self.hasher = argon2.PasswordHasher()

    def hash_password(self, plain_text: str) -> str:
        """Hash plain text password.

        :param plain_text: Password.
        :return: Hash of the password.
        """
        return self.hasher.hash(plain_text)

    def verify_password(self, hashed_password: str, plain_text: str) -> bool:
        """Validate if given hash and password match.

        :param hashed_password: Password hash.
        :param plain_text: Plain text password
        :return: Boolean indicating if plain_text relates to hashed_password.
        """
        try:
            self.hasher.verify(hashed_password, plain_text)
            verification = True
        except argon2.exceptions.VerifyMismatchError:
            verification = False
        return verification



"""HTTP request and response handling."""

# Pyramid
from pyramid.interfaces import ISession
from pyramid.registry import Registry
from pyramid.request import Request as _Request
from transaction import TransactionManager

# SQLAlchemy
from sqlalchemy.orm import Session

# Websauna
from websauna.system.admin.admin import Admin
from websauna.system.core.render import OnDemandResourceRenderer
from websauna.system.user.models import User


class Request(_Request):
    """HTTP request class.

    This is a Pyramid Request object augmented with type hinting information for Websauna-specific functionality.

    To know more about request read also

    * py:class:`pyramid.request.Request` documentation

    * py:class:`webob.request.Request` documentation

    Counter-intuitiveily this request is also available in non-HTTP applications like command line applications and timed tasks. These applications do not get request URL from a front end HTTP webserver, but a faux request is constructed pointing to the website URL taken from ``websauna.site_url`` setting. This is to allow similar design patterns and methodology to be applied in HTTP and non-HTTP applications.

    By settings variables in ``__type_hinting__()`` based on arguments types allows IDEs to infer type information when you hint your views as::

        from websauna.system.http import Request

        def hello_world(request:Request):
            request.  # <-- When typing, here autocompletion kicks in.

    """

    def __type_hinting__(self, user: User, dbsession: Session, session: ISession, admin: Admin, registry: Registry, on_demand_resource_renderer: OnDemandResourceRenderer, transaction_manager: TransactionManager):
        """A dummy helper function to tell IDEs about reify'ed variables.

        :param user: The logged in user. None if the visitor is anonymous.
        :param dbsession: Current active SQLAlchemy session
        :param session: Session data for anonymous and logged in users.
        :param admin: The default admin interface of the site. Note that the site can have several admin interfaces for different purposes.
        :param registry: Pyramid registry's. E.g. :py:attr:`pyramid.registry.Registry.settings` for reading settings and :py:meth:`pyramid.registry.Registry.notify` for sending events.
        :param on_demand_resource_renderer: Manage JS and CSS files which widgets want to pull on the page dynamically
        :param transaction_manager: Transaction manager used to commit the database changes after the request completes.
        """
        self.user = user
        self.dbsession = dbsession
        self.session = session
        self.admin = admin
        self.registry = registry
        self.on_demand_resource_renderer = on_demand_resource_renderer
        self.transaction_manager = transaction_manager
        self.tm = transaction_manager



# Standard Library
import sys
from codecs import open
from os import path

from setuptools import find_packages
from setuptools import setup


assert sys.version_info >= (3, 5, 2), "websauna.system needs Python 3.5.2 or newer, you have {version}".format(version=sys.version_info)

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='websauna.system',
    namespace_packages=["websauna"],
    version='1.0a8',
    description=long_description.split("\n")[0],
    long_description=long_description,
    url='https://websauna.org',
    author='Mikko Ohtamaa',
    author_email='mikko@opensourcehacker.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    keywords='sqlalchemy postgresql pyramid pytest websauna',
    packages=find_packages(exclude=['docs']),
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.5.2,<3.8',
    install_requires=[
        # Websauna dependencies
        'websauna.utils',

        # Pyramid dependencies
        'pyramid>=1.9',
        'waitress',
        'pyramid_redis_sessions',
        'pyramid-layout',
        "deform>=2.0.4",
        "pyramid_debugtoolbar",
        "pyramid_jinja2",

        # Time handling
        "arrow",
        "pytz",

        # SQLAlchemy and database support
        "alembic",
        "colanderalchemy",
        "psycopg2",
        "pyramid_retry",
        "pyramid_tm",
        "sqlalchemy",
        "sqlalchemy-utils",
        "zope.sqlalchemy",

        # User management
        "argon2_cffi",
        "authomatic",

        # Email
        'pyramid-mailer',
        'premailer',

        # Console logging
        "rainbow_logging_handler",

        # Misc
        "python-slugify",  # ASCII slug generation
    ],

    extras_require={
        # Dependencies needed to build and release Websauna
        'dev': [
            'ruamel.yaml',
            'setuptools_git',
            'pyroma==2.2',  # This is needed until version 2.4 of Pyroma is released
            'sphinx>=1.6.1',
            'sphinx-autodoc-typehints',
            'sphinx_rtd_theme',
            'sphinxcontrib-zopeext',
            'zest.releaser[recommended]'
        ],
        'test': [
            'celery',
            'websauna.tests',
            'cookiecutter',
            'codecov',
            'flake8',
            'pytest>=3.0',
            'pytest-runner',
            'coverage',
            'flaky',
            'isort',
            'pytest-cov',
            'pytest-runner',
            'pytest-splinter',
            'pytest-timeout',
            'webtest',
        ],
        "notebook": [
            "ipython[notebook]<5.2",
            "pyramid_notebook>=0.2.1",
            # Needed by python_notebook etc. who call pyramid.paster module
            "PasteDeploy",
        ],
        # Command line utilities and like that are needed to make development / production environment friendly
        'utils': ['pgcli'],
        # Using celery based async tasks
        'celery': ['celery[redis]>=4.1.0']
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'ws-sync-db=websauna.system.devop.scripts.syncdb:main',
            'ws-db-shell=websauna.system.devop.scripts.dbshell:main',
            'ws-shell=websauna.system.devop.scripts.shell:main',
            'ws-tweens=websauna.system.devop.scripts.tweens:main',
            'ws-alembic=websauna.system.devop.scripts.alembic:main',
            'ws-dump-db=websauna.system.devop.scripts.dumpdb:main',
            'ws-create-user=websauna.system.devop.scripts.createuser:main',
            'ws-celery=websauna.system.task.celeryloader:main',
            'ws-proutes=websauna.system.devop.scripts.proutes:main',
            'ws-pserve=websauna.system.devop.scripts.pserve:main',
            'ws-create-table=websauna.system.devop.scripts.createtable:main',
            'ws-sanity-check=websauna.system.devop.scripts.sanitycheck:main',
            'ws-collect-static=websauna.system.devop.scripts.collectstatic:main',
            'ws-settings=websauna.system.devop.scripts.settings:main',
        ],

        'paste.app_factory': [
            'main=websauna.system:main',

            # Scheduler automated test suite entry point with some extra configured taskss
            'task_test=websauna.tests.demotasks:main',
            'tutorial_test=websauna.tests.tutorial:main',
        ],
    },
)



# Standard Library
import inspect
import os
from types import ModuleType
from uuid import UUID

# SQLAlchemy
from sqlalchemy.ext.declarative import instrument_declarative


def secure_uuid():
    """Create a non-conforming 128-bit random version 4 UUID.

    Random UUID is a RFC 4122 compliant UUID version 4 128-bit number. By default 6 fixed bits, 4 bits for version and 2 bits reserved for other purposes, are fixed. This function behaves like Python's ` uuid4()`` but also randomizes the remaining six bits, generating up to 128 bit randomness.

    This function also sources all bytes from `os.urandom()` to guarantee the randomness and security and does not rely operating system libraries.

    Using ``secure_uuid()`` poses a risk that generated UUIDs are not accepted when communicating with third party system. However, they are observed to be good for URLs and to be stored in PostgreSQL.

    More information

    * http://crypto.stackexchange.com/a/3525/25874

    * https://tools.ietf.org/html/rfc4122
    """
    return UUID(bytes=os.urandom(16), version=4)


def attach_model_to_base(ModelClass: type, Base: type, ignore_reattach: bool=True):
    """Dynamically add a model to chosen SQLAlchemy Base class.

    More flexibility is gained by not inheriting from SQLAlchemy declarative base and instead plugging in models during the configuration time more.

    Directly inheriting from SQLAlchemy Base class has non-undoable side effects. All models automatically pollute SQLAlchemy namespace and may e.g. cause problems with conflicting table names. This also allows @declared_attr to access Pyramid registry.

    Example how to use Pyramid registry in SQLAlchemy's declared attributes::

        class SamplePluggableModel:
            '''Demostrate pluggable model which gets a referred class from Pyramid config.'''
            id = Column(Integer, primary_key=True)

            def __init__(self, **kwargs):
                # Default constructor
                self.__dict__.update(kwargs)

            @declared_attr
            def owner(cls):
                '''Refer to the configured user model.'''
                from websauna.system.user.utils import get_user_class
                config = cls.metadata.pyramid_config
                User = get_user_class(config.registry)
                return relationship(User, backref="referral_programs")

            @declared_attr
            def owner_id(cls):
                '''Refer to user.id column'''
                from websauna.system.user.utils import get_user_class
                config = cls.metadata.pyramid_config
                User = get_user_class(config.registry)
                return Column(Integer, ForeignKey('{}.id'.format(User.__tablename__)))

    Then you need to call in your Initializer::

        from example import models
        from websauna.system.model.meta import Base

        attach_model_to_base(models.SamplePluggableModel, Base)

    :param ModelClass: SQLAlchemy model class

    :param Base: SQLAlchemy declarative Base for which model should belong to

    :param ignore_reattach: Do nothing if ``ModelClass`` is already attached to base. Base registry is effectively global. ``attach_model_to_base()`` may be called several times within the same process during unit tests runs. Complain only if we try to attach a different base.
    """

    if ignore_reattach:
        if '_decl_class_registry' in ModelClass.__dict__:
            assert ModelClass._decl_class_registry == Base._decl_class_registry, "Tried to attach to a different Base"
            return

    instrument_declarative(ModelClass, Base._decl_class_registry, Base.metadata)

    # TODO: We now hardcode this event listener here.
    # from sqlalchemy.orm.instrumentation import _instrumentation_factory
    # _instrumentation_factory.dispatch.class_instrument(ModelClass)

    # The correct approach is through class_instrument() event firing,
    # but could not figure out yet how to make it happen with all bits flying around
    from .json import init_for_json
    init_for_json(ModelClass)


def attach_models_to_base_from_module(mod: ModuleType, Base: type):
    """Attach all models in a Python module to SQLAlchemy base class.

    The attachable models must declare ``__tablename__`` property and must not have existing ``Base`` class in their inheritance.
    """

    for key in dir(mod):
        value = getattr(mod, key)
        if inspect.isclass(value):

            # TODO: We can't really check for SQLAlchemy declarative base class as it's usually run-time generated and may be out of our control
            if any(base.__name__ == "Base" for base in inspect.getmro(value)):
                # Already inhertis from SQALchemy declarative Base
                continue

            if hasattr(value, "__tablename__"):
                # This declares table but is not attached to any base yet
                attach_model_to_base(value, Base)



"""Password reset."""
# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response
from zope.interface import implementer

# Websauna
from websauna.system.core import messages
from websauna.system.core.route import get_config_route
from websauna.system.http import Request
from websauna.system.mail import send_templated_mail
from websauna.system.user.events import PasswordResetEvent
from websauna.system.user.events import UserAuthSensitiveOperation
from websauna.system.user.interfaces import CannotResetPasswordException
from websauna.system.user.interfaces import ICredentialActivityService
from websauna.system.user.interfaces import IUser
from websauna.system.user.utils import get_user_registry


@implementer(ICredentialActivityService)
class DefaultCredentialActivityService:
    """Handle password reset process and such."""

    def __init__(self, request: Request):
        self.request = request

    def create_forgot_password_request(self, email: str, location: str=None) -> Response:
        """Create a new email activation token for a user and produce the following screen.

        * Sets user password reset token
        * Sends out reset password email
        * The existing of user with such email should be validated beforehand

        :param email: User email.
        :param location: URL to redirect the user after the password request.
        :return: Redirect to location.
        :raise: CannotResetPasswordException if there is any reason the password cannot be reset. Usually wrong email.
        """
        request = self.request

        user_registry = get_user_registry(request)

        reset_info = user_registry.create_password_reset_token(email)
        if not reset_info:
            raise CannotResetPasswordException("Cannot reset password for email: {email}".format(email=email))
        user, token, expiration_seconds = reset_info

        link = request.route_url('reset_password', code=token)
        context = dict(link=link, user=user, expiration_hours=int(expiration_seconds / 3600))
        send_templated_mail(request, [email, ], "login/email/forgot_password", context=context)

        messages.add(request, msg="Please check your email to continue password reset.", kind='success', msg_id="msg-check-email")

        if not location:
            location = get_config_route(request, 'websauna.request_password_reset_redirect')
            assert location

        return HTTPFound(location=location)

    def get_user_for_password_reset_token(self, activation_code: str) -> IUser:
        """Get a user by activation token.

        :param activation_code: User activation code.
        :return: User for the given activation_code.
        """
        request = self.request
        user_registry = get_user_registry(request)
        user = user_registry.get_user_by_password_reset_token(activation_code)
        return user

    def reset_password(self, activation_code: str, password: str, location: str=None) -> Response:
        """Perform actual password reset operations.

        User has following password reset link (GET) or enters the code on a form.

        :param activation_code: Activation code provided by the user.
        :param password: New user password.
        :param location: URL to redirect the user after the password request.
        :return: Redirect to location.
        :raise: HTTPNotFound if activation_code is not found.
        """
        request = self.request
        user_registry = get_user_registry(request)
        user = user_registry.get_user_by_password_reset_token(activation_code)
        if not user:
            return HTTPNotFound("Activation code not found")

        user_registry.reset_password(user, password)

        messages.add(request, msg="The password reset complete. Please sign in with your new password.", kind='success', msg_id="msg-password-reset-complete")

        request.registry.notify(PasswordResetEvent(self.request, user, password))
        request.registry.notify(UserAuthSensitiveOperation(self.request, user, "password_reset"), request)

        location = location or get_config_route(request, 'websauna.reset_password_redirect')
        return HTTPFound(location=location)



# Standard Library
import os

import pytest

# Websauna
from websauna.system import Initializer
from websauna.system.core.route import add_template_only_view
from websauna.tests.fixtures import get_app
from websauna.tests.webserver import customized_web_server


HERE = os.path.abspath(os.path.dirname(__file__))


def extra_init(init: Initializer):
    """Configure one templated only view."""
    config = init.config
    config.add_jinja2_search_path(HERE + "/templates", name=".html")
    add_template_only_view(config, "/dummy", "dummy", "dummy.html")


@pytest.fixture(scope="module")
def app(request, ini_settings):
    """Construct a WSGI app with tutorial models and admins loaded."""
    app = get_app(ini_settings, extra_init=extra_init)
    return app


@pytest.fixture(scope="module")
def web_server(request, app):
    """Run a web server
    with tutorial installed."""
    web_server = customized_web_server(request, app)
    return web_server()


def test_template_only_view(browser, web_server):
    """See that we can register and render a template only view."""

    browser.visit(web_server + "/dummy")



"""Core events."""

# Websauna
from websauna.system.http import Request


class InternalServerError:
    """Fired when internal server error page is rendered.

    Event can be consumed by monitoring addons like websauna.sentry.
    """

    def __init__(self, exception: Exception, request: Request):
        self.exception = exception
        self.request = request



"""Adaption of tomb_routes.

This code is adoption of tomb_routes https://github.com/TombProject/tomb_routes created by John Anderson.
"""

# Pyramid
from pyramid.interfaces import IViewMapperFactory
from pyramid.path import DottedNameResolver


def add_simple_route(
        config, path, target,
        append_slash=False,
        *args, **kwargs
):
    """Configuration directive that can be used to register a simple route to
    a view.

    Examples:

    with view callable::

        config.add_simple_route(
            '/path/to/view', view_callable,
            renderer='json'
        )

    with dotted path to view callable::

        config.add_simple_route(
            '/path/to/view', 'dotted.path.to.view_callable',
            renderer='json'
        )
    """

    target = DottedNameResolver().maybe_resolve(target)
    mapper = config.get_routes_mapper()

    route_name = kwargs.pop("route_name", None)
    route_name = route_name or target.__name__
    route_name_count = 0

    # Arguments passed to route
    route_kwargs = {}

    if 'accept' in kwargs:
        val = kwargs.pop('accept')
        route_kwargs['accept'] = val

    # Make it possible to custom_predicates = in the simple_route
    custom_predicates = kwargs.pop('custom_predicates', None)
    if custom_predicates:
        route_kwargs["custom_predicates"] = custom_predicates

    if 'attr' in kwargs:
        route_name += '.' + kwargs['attr']

    routes = {route.name: route for route in mapper.get_routes()}
    orig_route_name = route_name

    while route_name in routes:
        route_name = '%s_%s' % (orig_route_name, route_name_count)
        route_name_count += 1

    current_pregen = kwargs.pop('pregenerator', None)

    def pregen(request, elements, kwargs):
        if 'optional_slash' not in kwargs:
            kwargs['optional_slash'] = ''

        if current_pregen is not None:
            return current_pregen(request, elements, kwargs)
        else:
            return elements, kwargs

    orig_route_prefix = config.route_prefix
    # We are nested with a route_prefix but are trying to
    # register a default route, so clear the route prefix
    # and register the route there.
    if (path == '/' or path == '') and config.route_prefix:
        path = config.route_prefix
        config.route_prefix = ''

    if append_slash:
        path += '{optional_slash:/?}'
        config.add_route(
            route_name, path, pregenerator=pregen,
            **route_kwargs
        )
    else:
        config.add_route(
            route_name, path, pregenerator=current_pregen,
            **route_kwargs
        )

    kwargs['route_name'] = route_name

    if 'mapper' not in kwargs:
        # This should default to 'websauna.system.core.csrf.csrf_mapper_factory.<locals>.CSRFMapper'>
        mapper = config.registry.queryUtility(IViewMapperFactory)
        kwargs['mapper'] = mapper

    config.add_view(target, *args, **kwargs)
    config.commit()
    config.route_prefix = orig_route_prefix


def includeme(config):
    """Function that gets called when client code calls config.include"""
    config.add_directive('add_simple_route', add_simple_route)



from .interfaces import ISecrets


def get_secrets(registry) -> dict:
    """Get the secrets provider dictionary.

    :return: A dictionary containing secrets having {ini sectionid}.{key} as keys.
    """
    return registry.getUtility(ISecrets)



from pyramid_mailer.mailer import DummyMailer


class StdoutMailer:
    """Print all outgoing email to console.

    Used by the development server.
    """

    def __init__(self):
        # For testing purposes
        self.send_count = 0

    def _send(self, message, fail_silently=False):
        """Save message to a file for debugging
        """
        print(str(message.to_message()))
        self.send_count += 1

    send = _send
    send_immediately = _send
    send_to_queue = _send
    send_sendmail = _send
    send_immediately_sendmail = _send


class NullMailer:
    """Ignore all otugoing mail and only increase send count.

    Functional testing email backend.
    """

    def __init__(self):
        # For testing purposes
        self.send_count = 0

    def _send(self, message, fail_silently=False):
        """Save message to a file for debugging
        """
        self.send_count += 1

    send = _send
    send_immediately = _send
    send_to_queue = _send
    send_sendmail = _send
    send_immediately_sendmail = _send


class ThreadFriendlyDummyMailer(DummyMailer):
    """Multi-thread aware mailing test backend.

    We store outbox messages in class globals. If a web server or another thread sends out a message this allows us to access the message in a test thread.
    """

    outbox = []
    queue = []

    def __init__(self):
        # Override DummyMailer self init here
        pass

    @classmethod
    def reset(cls):
        ThreadFriendlyDummyMailer.outbox = []
        ThreadFriendlyDummyMailer.queue = []

    def _send(self, message, fail_silently=False):
        """Save message to a file for debugging"""
        self.output.append(message)
        self.send_count += 1



# Pyramid
import pyramid.testing
from pyramid.testing import DummyRequest

import pytest

# Websauna
import websauna.system
from websauna.system.core import sitemap
from websauna.system.core.sitemap import ReflectiveSitemapBuilder
from websauna.system.core.sitemap import RouteItem
from websauna.system.core.sitemap import TraverseItem
from websauna.system.http.utils import make_routable_request


@pytest.fixture(scope="module")
def sitemap_app(request, paster_config):
    '''Custom WSGI app with travesal points for sitemap enabled.'''

    class Initializer(websauna.system.DemoInitializer):

        def configure_views(self):
            super(Initializer, self).configure_views()
            from websauna.tests import sitemapsamples
            self.config.add_route('sitemap_test', '/container/*traverse', factory=sitemapsamples.container_factory)
            self.config.scan(sitemapsamples)

    global_config, app_settings = paster_config
    init = Initializer(global_config, app_settings)
    init.run()
    app = init.make_wsgi_app()
    app.init = init
    return app


@pytest.fixture
def sitemap_request(sitemap_app):
    """Create a dummy request object useable for sitemap building tests."""
    return make_routable_request(dbsession=None, registry=sitemap_app.init.config.registry)


@pytest.fixture
def builder(sitemap_request):
    return ReflectiveSitemapBuilder(sitemap_request)


def test_route_items():
    """Routed items appear in the sitemap."""

    with pyramid.testing.testConfig() as config:

        request = DummyRequest()
        config.add_route('bar', '/bar/{id}')

        s = sitemap.Sitemap()
        s.add_item(sitemap.RouteItem("bar", id=1, lastmod="2015-01-01"))
        s.add_item(sitemap.RouteItem("bar", id=2, priority="1.0"))
        s.add_item(sitemap.RouteItem("bar", id=3, changefreq="never"))

        data = s.render(None, request)
        items = list(data["urlset"])
        assert len(items) == 3
        assert items[0].location(request) == "http://example.com/bar/1"
        assert items[0].lastmod(request) == "2015-01-01"

        assert items[1].location(request) == "http://example.com/bar/2"
        assert items[1].priority(request) == "1.0"

        assert items[2].location(request) == "http://example.com/bar/3"
        assert items[2].changefreq(request) == "never"


def test_generator_items():
    """Generated items appear in the sitemap."""

    class FooItem(sitemap.SitemapItem):

        def __init__(self, name):
            super(FooItem, self).__init__()
            self.name = name

        def location(self, request):
            return "/" + str(self.name)

    def generator():
        items = [FooItem(1), FooItem(2), FooItem(3)]
        yield from items

    with pyramid.testing.testConfig():
        request = DummyRequest()
        s = sitemap.Sitemap()
        s.add_generator(generator)
        data = s.render(None, request)
        items = list(data["urlset"])
        assert len(items) == 3
        assert items[0].location(request) == "/1"
        assert items[1].location(request) == "/2"
        assert items[2].location(request) == "/3"


def test_reflect_routes(builder):
    """See we can reflect simple routes back to the sitemap."""

    builder.build_routes()
    routes = [i.route_name for i in builder.sitemap.items if isinstance(i, RouteItem)]

    assert "parameter_free_route" in routes
    assert "permissioned_route" not in routes
    assert "post_only_route" not in routes


def test_reflect_traverse(builder):
    """See we can reflect traverse hierarchy back to the sitemap."""
    builder.build_traverse_trees()
    request = builder.request
    urls = [i.location(request) for i in builder.sitemap.items if isinstance(i, TraverseItem)]

    # Admin traversing is protected by view permission
    assert not(any("admin" in u for u in urls))

    # We get default view (empty name) and named view
    assert any(u.endswith("/container/") for u in urls)
    assert any(u.endswith("/container/additional") for u in urls)

    # See we nest correctly
    assert any("/nested/foo/" in u for u in urls)
    assert any("/nested/bar/" in u for u in urls)
    assert any("/nested/bar/additional" in u for u in urls)

    # See we dont' grab permissioned views
    assert not (any("permissioned" in u for u in urls))


def test_reflect_build(builder):
    """Build both routes and traversables."""
    builder.build()


def test_conditions(builder):
    """We can enable/disable items in the sitemap using decorators."""

    builder.build_routes()
    routes = [i.route_name for i in builder.sitemap.items if isinstance(i, RouteItem)]
    assert "conditional_route" in routes
    assert "skipped_route" not in routes

    builder.build_traverse_trees()
    request = builder.request
    urls = [i.location(request) for i in builder.sitemap.items if isinstance(i, TraverseItem)]

    assert any("/container/bar/conditional" in u for u in urls)
    assert not any("/container/bar/skipped_conditional" in u for u in urls)



"""ws-shell script.

IPython shell prompt to Websauna.
"""
# Standard Library
import datetime
import sys
import typing as t
from collections import OrderedDict

# Pyramid
import transaction

# Websauna
from websauna.system.core.redis import get_redis
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import feedback
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message
from websauna.system.model.meta import Base
from websauna.utils.time import now


try:
    from IPython import embed
except ImportError as e:
    raise ImportError("You need to install IPython to use this shell") from e


def main(argv: t.List[str]=sys.argv):
    """Execute the IPython shell prompt with Websauna configuration already initialised.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv, additional_params='[var=value]')

    config_uri = get_config_uri(argv)

    request = init_websauna(config_uri)

    imported_objects = OrderedDict()

    imported_objects["request"] = request
    imported_objects["dbsession"] = request.dbsession
    imported_objects["transaction"] = transaction
    imported_objects["redis"] = get_redis(request)
    imported_objects["now"] = now
    imported_objects["datetime"] = datetime

    for name, cls in Base._decl_class_registry.items():

        if name == "_sa_module_registry":
            continue

        imported_objects[name] = cls

    feedback('', False)
    feedback('Following classes and objects are available:', False)
    for var, val in imported_objects.items():
        line = "{key:30}: {value}".format(
            key=var,
            value=str(val).replace('\n', ' ').replace('\r', ' ')
        )
        feedback(line)
    feedback('', False)

    embed(user_ns=imported_objects)


if __name__ == "__main__":
    main()



"""Timed default devop tasks for the system."""
# Standard Library
import logging

# Websauna
from websauna.system.task.tasks import task


logger = logging.getLogger(__name__)


@task(name="backup")
def backup_task():
    from . import backup
    logger.info("Running daily backup")
    backup.backup_site()
    logger.info("Daily backup done")



"""SQLAlchemy integration for Colander and Deform frameworks."""
# Standard Library
import typing as t

# Pyramid
import colander
from colander.compat import is_nonstr_iter

# SQLAlchemy
from sqlalchemy import Column
from sqlalchemy.orm import Query
from sqlalchemy.orm import Session

# Websauna
from websauna.utils.slug import slug_to_uuid
from websauna.utils.slug import uuid_to_slug


def extract_uuid_to_slug(item):
    """Reads uuid attribute on the model name returns it as B64 encoded slug."""
    return uuid_to_slug(item.uuid)


def convert_query_to_tuples(query: Query, first_column: t.Union[str, t.Callable], second_column: t.Union[str, t.Callable], default_choice: t.Optional[str] =None) -> t.List[t.Tuple[str, str]]:
    """Convert SQLAlchemy query results to (id, name) tuples for select and checkbox widgets.

    :param first_column: Column name used to populate value in the first tuple
    :param second_column: Column name used to populate value in the second tuple
    :oaram default_choice: If given use this as "Select here" or when the value is None
    """
    if type(first_column) == str:
        first_column_getter = lambda item: getattr(item, first_column)
    else:
        first_column_getter = first_column

    if type(second_column) == str:
        second_column_getter = lambda item: getattr(item, second_column)
    else:
        second_column_getter = second_column

    result = []

    if default_choice:
        result.append(('', default_choice))

    for item in query:
        result.append((first_column_getter(item), second_column_getter(item)))

    return result


def get_uuid_vocabulary_for_model(dbsession: Session, model: type, first_column=extract_uuid_to_slug, second_column=str, default_choice=None) -> t.List[t.Tuple]:
    """Create a select/checkbox vocabulary containing all items of a model."""
    query = dbsession.query(model).all()
    return convert_query_to_tuples(query, first_column, second_column, default_choice)


class ModelSetResultList(list):
    """Mark that the result is through SQLAlchemy query."""


class ModelSchemaType:
    """Mixin class for using SQLAlchemy with Colander types.

    Provide utilitiy functions to extract model and dbsession from the context.
    """

    #: Point this to the model this set is supposed to query
    model = None

    #: Name of the column on the model we use to fetch objects using IN query
    match_column = None

    #: Name of the column which provides label or such for items in sequence. If not present item __str__ is used.
    label_column = None

    def __init__(self, model: type=None):
        if model:
            self.model = model

    def get_dbsession(self, node) -> Session:
        return node.bindings["request"].dbsession

    def get_model(self, node) -> type:
        """Which model we are quering."""
        return self.model

    def convert_to_id(self, item):
        id = getattr(item, self.match_column)
        value = getattr(item, self.label_column)
        return (id, value)

    def get_match_column(self, node: colander.SchemaNode, model: type) -> Column:
        """Get the column we are filtering out."""
        assert self.match_column, "match_column undefined"
        return getattr(model, self.match_column)


class ForeignKeyValue(ModelSchemaType, colander.String):
    """Hold a reference to one SQLAlchemy object for Colander schema serialization.

    See :ref:`Widgets <deform:widget>` for more information.
    """

    def serialize(self, node, appstruct):

        if appstruct is colander.null:
            return colander.null

        value = self.preprocess_appstruct_value(node, appstruct)
        return value

    def preprocess_cstruct_value(self, node: colander.SchemaNode, cstruct: set) -> t.Union[set, t.List]:
        """Parse incoming form values to Python objects if needed.
        """
        return cstruct

    def preprocess_appstruct_value(self, node: colander.SchemaNode, appstruct: set) -> t.List[str]:
        """Convert items to appstruct ids.
        """
        return str(appstruct)

    def query_item(self, node: colander.SchemaNode, dbsession: Session, model: type, match_column: Column, value: set) -> t.List[object]:
        """Query the actual model to get the concrete SQLAlchemy objects."""

        if not value:
            # Empty IN queries are not allowed
            return value

        return dbsession.query(model).filter(match_column == value).first()

    def deserialize(self, node, cstruct: str):
        """Convert incoming form value - string id - back to a SQLAlchemy object."""
        if cstruct is colander.null:
            return colander.null

        dbsession = self.get_dbsession(node)
        model = self.get_model(node)
        match_column = self.get_match_column(node, model)
        value = self.preprocess_cstruct_value(node, cstruct)
        value = self.query_item(node, dbsession, model, match_column, value)

        return value


class ModelSet(ModelSchemaType, colander.Set):
    """Presents set of chosen SQLAlchemy models instances.

    This automatically turns SQLAlchemy objects to (id, label) tuples, so that they can be referred in various widgets (select, checkbox).

    See :ref:`Widgets <deform:widget>` for more information.
    """

    def serialize(self, node, appstruct):

        assert self.match_column, "match_column not configured"

        if appstruct is colander.null:
            return colander.null

        values = self.preprocess_appstruct_values(node, appstruct)
        return values

    def deserialize_set_to_models(self, node, cstruct):
        dbsession = self.get_dbsession(node)
        model = self.get_model(node)
        match_column = self.get_match_column(node, model)
        values = self.preprocess_cstruct_values(node, cstruct)
        return self.query_items(node, dbsession, model, match_column, values)

    def preprocess_cstruct_values(self, node: colander.SchemaNode, cstruct: set) -> t.Union[set, t.List]:
        """Parse incoming form values to Python objects if needed.
        """
        return cstruct

    def preprocess_appstruct_values(self, node: colander.SchemaNode, appstruct: set) -> t.List[str]:
        """Convert items to appstruct ids.
        """
        if self.label_column:
            return [getattr(i, self.label_column) for i in appstruct]
        else:
            return [str(i) for i in appstruct]

    def query_items(self, node: colander.SchemaNode, dbsession: Session, model: type, match_column: Column, values: set) -> t.List[object]:
        """Query the actual model to get the concrete SQLAlchemy objects."""
        if not values:
            # Empty IN queries are not allowed
            return []
        return ModelSetResultList(dbsession.query(model).filter(match_column.in_(values)).all())

    def deserialize(self, node, cstruct):

        if cstruct is colander.null:
            return colander.null

        if not is_nonstr_iter(cstruct):
            raise colander.Invalid(node, '{} is not iterable'.format(cstruct))

        return self.deserialize_set_to_models(node, cstruct)


class UUIDForeignKeyValue(ForeignKeyValue):
    """Hold a reference to SQLAlchemy object through base64 encoded UUID value.

    Useful for select widget, etc.

    See :ref:`Widgets <deform:widget>` for more information.
    """

    #: The name of the column from where we extract UUID value. Defauts to ``uuid``.
    match_column = "uuid"

    def __init__(self, model, match_column=None):
        super().__init__(model)
        if match_column:
            self.match_column = match_column

    def preprocess_cstruct_value(self, node, cstruct):
        """Parse incoming form values to Python objects if needed.
        """
        return slug_to_uuid(cstruct)

    def preprocess_appstruct_value(self, node: colander.SchemaNode, appstruct: set) -> t.List[str]:
        """Convert items to appstruct ids.
        """
        return uuid_to_slug(getattr(appstruct, self.match_column))


class UUIDModelSet(ModelSet):
    """A set of SQLAlchemy objects queried by base64 encoded UUID value.

    See :ref:`Widgets <deform:widget>` for more information.
    """

    match_column = "uuid"

    def __init__(self, model: type=None, match_column: str=None):
        if model:
            self.model = model

        if match_column:
            self.match_column = match_column

    def preprocess_cstruct_values(self, node, cstruct):
        """Parse incoming form values to Python objects if needed.
        """
        return [slug_to_uuid(v) for v in cstruct]

    def preprocess_appstruct_values(self, node: colander.SchemaNode, appstruct: set) -> t.List[str]:
        """Convert items to appstruct ids.
        """
        return [uuid_to_slug(getattr(i, self.match_column)) for i in appstruct]



"""Redis connection manager."""
# Standard Library
import logging
import os
import threading
import typing as t

# Pyramid
from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.events import subscriber
from pyramid.registry import Registry

from redis import ConnectionError
from redis import ConnectionPool
from redis import StrictRedis

# Websauna
from websauna.system.http import Request


logger = logging.getLogger(__name__)


def create_redis(registry: Registry, connection_url=None, redis_client=StrictRedis, max_connections=16, **redis_options) -> StrictRedis:
    """Sets up Redis connection pool once at the start of a process.

    Connection pool life cycle is the same as Pyramid registry which is the life cycle of a process (all threads).

    :param max_connections: Default per-process connection pool limit
    """

    # if no url passed, try to get it from pyramid settings
    url = registry.settings.get('redis.sessions.url') if connection_url is None else connection_url
    # otherwise create a new connection
    if url is not None:
        # remove defaults to avoid duplicating settings in the `url`
        redis_options.pop('password', None)
        redis_options.pop('host', None)
        redis_options.pop('port', None)
        redis_options.pop('db', None)
        # the StrictRedis.from_url option no longer takes a socket
        # argument. instead, sockets should be encoded in the URL if
        # used. example:
        #     unix://[:password]@/path/to/socket.sock?db=0
        redis_options.pop('unix_socket_path', None)

        # connection pools are also no longer a valid option for
        # loading via URL
        redis_options.pop('connection_pool', None)

        process_name = os.getpid()
        thread_name = threading.current_thread().name

        logger.info("Creating a new Redis connection pool. Process %s, thread %s, max_connections %d", process_name, thread_name, max_connections)

        connection_pool = ConnectionPool.from_url(url, max_connections=max_connections, **redis_options)
        redis = StrictRedis(connection_pool=connection_pool)
    else:
        raise RuntimeError("Redis connection options missing. Please configure redis.sessions.url")

    return redis


def log_redis_statistics(redis: StrictRedis):
    """Log Redis connection pool statistics at the end of each request.

    Help to diagnose connection problems and web process lifetime issues.

    We do not need to do specific cleanup, as connnection pool *should* clean up the connections when they go out of scope. Each Redis client function calls connection.release() at the end of operation."
    """
    pool = redis.connection_pool

    max_connections = pool.max_connections
    created = pool._created_connections
    available = len(pool._available_connections)
    in_use = len(pool._in_use_connections)

    process_name = os.getpid()
    thread_name = threading.current_thread().name

    logger.debug("Redis connection statistics - process: %s, thread: %s, created: %d, max: %d, in-use: %d, available: %d", process_name, thread_name, created, max_connections, available, in_use)


def get_redis(request_or_registry: t.Union[Request, Registry], url: str=None, redis_client=StrictRedis, **redis_options) -> StrictRedis:
    """Get a connection to Redis.

    Example:

    .. code-block:: python

        from websauna.system.core.redis import get_redis

        def my_view(request):
            redis = get_redis(request)
            redis.set("myval", "foobar")
            print(redis.get("myval))

    See :ref:`transient` data documentation.

    `See Redis command documentation <http://redis.io/commands>`_.

    `See Redis Python client <https://pypi.python.org/pypi/redis>`_.

    Compatible with *pyramid_redis_session*, see https://github.com/ericrasmussen/pyramid_redis_sessions/blob/master/pyramid_redis_sessions/connection.py

    :param request: HTTP request object. NOTE: By legacy this argument also supports Registry object. However this behaviro will be deprecated.

    :param url: An optional connection string that will be passed straight to `StrictRedis.from_url`. The connection string should be in the form redis://username:password@localhost:6379/0. If not given use the default from settings.

    :param redis_options: A dict of keyword args to be passed straight to `StrictRedis`

    :return: Redis client
    """
    # TODO: Resolve calling convention only using one
    if isinstance(request_or_registry, Registry):
        # Unit test calling convention
        registry = request_or_registry
        # request = None
    else:
        registry = request_or_registry.registry
        # request = request_or_registry

    redis = registry.redis
    return redis


def is_sane_redis(config: Configurator) -> bool:
    """Check that we have a working Redis connection for session.

    Execute this on startup, so we bail out without starting up with a missing Redis.

    :return: True if Redis connection works
    """
    try:
        redis = get_redis(config.registry)
        redis.set("websauna_session_test", True)
        return True
    except ConnectionError as e:
        return False


@subscriber(NewRequest)
def on_request_log_redis_stats(event):
    """On every request dump Redis diagnostics information."""
    redis = get_redis(event.request)
    log_redis_statistics(redis)



"""Register UserEvent subscribers."""
# Pyramid
from pyramid.events import subscriber

# Websauna
from websauna.system.user.events import UserCreated
from websauna.system.user.utils import get_site_creator


@subscriber(UserCreated)
def site_init(e: UserCreated):
    """Initialize the website.

    When the first user hits the site, capture its login and add him to the admin group.

    :param e: UserCreated event.
    """
    request = e.request
    user = e.user
    site_creator = get_site_creator(request.registry)
    site_creator.check_empty_site_init(request.dbsession, user)



"""ws-settings script.


Display settings for a given configuration file.
"""
# Standard Library
import sys
import typing as t
from pprint import pprint

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import feedback
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message


def main(argv: t.List[str]=sys.argv):
    """Display settings for a given configuration file.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    config_uri = get_config_uri(argv)

    request = init_websauna(config_uri, sanity_check=False)
    message = 'Active deployment settings in {config_uri}'.format(config_uri=config_uri)
    feedback(message)
    pprint(request.registry.settings)
    feedback_and_exit('', status_code=0, display_border=True)



"""Defining redirects quickly."""

# Pyramid
import venusian
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPMovedPermanently

from slugify import slugify

# Websauna
from websauna.system.core.simpleroute import add_simple_route


def redirect_view(path: str, new_path: str=None, new_route: str=None, status_code: int=302, name: str=None):
    """Generate a redirect view.

    Assign the return value of this function to a module level variable. `configuration.scan()` will pick it up and make a redirect route.

    Give a new route name or direct path where requests should be redirected.

    Example usage:

    .. code-block:: python

        # Product no longer available
        _redirect = redirect_view("/old-page", new_route="home", status_code=301)
        _redirect2 = redirect_view("/old-page-2", new_route="home", status_code=301)

        # SEO optimized path
        _seo_redirect = redirect_view("/old-path-name", new_route="new_route_name", status_code=301)

    TODO:

    * Add regex matching and such.

    ' Inefficient internal implementation

    :param path: Old path to be redirected to a new location

    :param new_route: A named route where requests are redirected.

    :param new_path: A URL path where requests are redirected.

    :param status_code: 301 for permanent redirect, 302 for temporary direct

    :param name: Optional route name. If not set it is generated from path.
    """

    assert new_route or new_path

    if not name:
        name = slugify(path) + "_redirect"

    new_path = new_path

    def redirect_view(request):

        if new_path:
            mapped_path = new_path
        else:
            mapped_path = request.route_url(new_route)

        if status_code == 302:
            return HTTPFound(mapped_path)
        elif status_code == 301:
            return HTTPMovedPermanently(mapped_path)
        else:
            raise RuntimeError("Unsupported redirect code {}", status_code)

    def callback(scanner, _name, wrapped):
        """Register a view; called on config.scan"""
        config = scanner.config

        # pylint: disable=W0142
        add_simple_route(config, path, redirect_view, append_slash=False, route_name=name)

    venusian.attach(redirect_view, callback)

    return redirect_view
